//<: 'dir.c' v0.046 @ 2012-10-08 20:05:24 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS global directory functions */

#include <stdlib.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <dirent.h>

#include "knobs.h"
#include "global.h"
#include "dnsdata.h"
#include "strutil.h"
#include "dir.h"
#include "log.h"

#define MAX_PATH_LEN 1024

// We save some data when looking up records, so it may be reused faster.

// Actual name for last found record:
char found_name[MAX_RECORD_LEN + 1] = "";
// Labels cut of from name to find it:
int found_cuts = 0;
// Saved time of last found record:
time_t found_mtime = 0;
// Saved record types of last found record:
int searched_qtype = 0;
int found_qtype = 0;
int got_qtype = 0;
// Path of last record searched for:
char rr_path[MAX_PATH_LEN+1] = "";


// get hash for string
char *hex_hash(unsigned char *s)
{
  static char h[4];
  unsigned long r = 0;
  for (; *s; s++)
    r += *s;
  switch(hash_type)
  {
    case HASH_NEW:
      snprintf (h, sizeof (h), "#%2.2X", (unsigned int)(r & 0xFFUL));
      break;
    case HASH_OLD:
      snprintf (h, sizeof (h), "%2.2X", (unsigned int)(r & 0xFFUL));
      break;
    default:
      h[0] = '!';
      h[1] = 0; 
  }
  return h;
}

// get all hashes
char **hex_hashes(int type)
{
  if (type == HASH_NO)
    return NULL;
  const char *hex = "0123456789ABCDEF";
  int il = 256 * sizeof(char *);
  char **hl = malloc(il + (256 * 4));
  hl[il] = NULL;
  char *hp = (char *)&(hl[il+1]);
  for (int i = 0; i < 256; i++)
  {
    hl[i] = hp;
    if (type == HASH_NEW)
      *(hp++) = '#';
    *(hp++) = hex[i >> 4];
    *(hp++) = hex[i & 0xF];
    *(hp++) = '\0';
  }
  return hl;
}

int not_db_name(char *name, int flags)
{
  if (!name)
    return -1;
  if (!(flags & SDH_INC_PSEUDO))
  {
    if (bad_db_name_first_char(*name))
      return 1;
    for (int i = 1; name[i]; i++)
      if bad_db_name_char(name[i])
        return i+1;
  } else {
    if ((*name != '!') && (*name != '~') && bad_db_name_first_char(*name))
      return 1;
    for (int i = 1; name[i]; i++)
      if ((*name != '!') && (*name != '~') && bad_db_name_char(name[i]))
        return i+1;
  }
  return 0;
}


// Get directory for domain name
int get_dir_path(char *buf, int len, char *name)
{
  if (len < 1)
    return 0;
  if (!name)
  {
    strncpy(buf, data_dir, len);
    return strlen(data_dir);
  }
  if (hash_type == HASH_NO)
    return snprintf(buf, len, "%s/%s", data_dir, name);
  return snprintf(buf, len, "%s/%s/%s", data_dir, hex_hash((unsigned char *)name), name);
}

// Get file path for query
int get_file_path(char *buf, int len, char *name, char *type)
{
  if (len < 1)
    return 0;
  if (!type)
    return get_dir_path(buf, len, name);
  if (!name)
    return snprintf(buf, len, "%s/%s", data_dir, type);
  if (hash_type == HASH_NO)
    return snprintf(buf, len, "%s/%s/%s", data_dir, name, type);
  return snprintf(buf, len, "%s/%s/%s/%s", data_dir, hex_hash((unsigned char *)name), name, type);
}

// Create a directory name directory
void make_directory(char *dir, char* desc)
{
  char buf[1024];
  if (get_dir_path(buf, sizeof(buf), dir) && !mkdir(buf, NEW_DIR_MODE))
  {
    if (desc)
      logmsg(LOG_INFO, "created %s directory: %s", desc, buf);
    else
      logmsg(LOG_INFO, "created %s directory: %s", dir, buf);
  }
}

// Create a RR file and write contents to it
void make_file(char *dir, char *type, char *data, char *desc)
{
  char buf[1024];
  if (get_file_path(buf, sizeof(buf), dir, type))
  {
    int fd = open(buf, O_WRONLY | O_CREAT | O_EXCL, NEW_FILE_MODE);
    if (fd != -1)
    {
      int rw = write(fd, data, strlen(data));
      int rc = close(fd);
      if (rc || (rw != strlen(data)))
        logmsg(LOG_WARN, "failed to create file: %s %s", dir, type);
      else if (desc)
        logmsg(LOG_INFO, "created %s file: %s", desc, buf);
      else
        logmsg(LOG_INFO, "created %s %s file: %s", dir, type, buf);
    }
  }
}

// Get path to cached record instead of normal record:
//int get_cache_path(char *buf, int len, char *path)
//{
//  int l = len;
//  i = strlen(data_dir);
//  if (strncmp(path, data_dir, i))
//    return 0;
//  strncpy(buf, path, min(i,l));
//  if (l <= i)
//    return 0;
//  l -= i;
//  strncpy(buf+i, "/:cache", min(l,7));
//  if (l <= 7)
//    return 0;
//  l -= 7;
//  strncpy(buf+i+7, path+i, l);
//  l -= strlen(path+i);
//  return (l < 0) ? 0 : len - l;
//}

// Get path to record related to previous record:
int get_related_path(char *path, int len, char *orgpath, int baseqtype, int relqtype)
{
  if (orgpath)
  {
    if (!*orgpath)
      return 0;
    strncpy(path, orgpath, len);
    if (path[len-1])
      return 0;
  }
  if (!baseqtype && found_qtype)
    baseqtype = searched_qtype;
  char *baseqname = qtype_fname(baseqtype);
  if (!(baseqname && *baseqname))
    return 0;
  char *relqname = NULL;
  if (relqtype)
  {
    relqname = qtype_fname(relqtype);  
    if (!(relqname && *relqname))
      return 0;
  }
  int i = strlen(path);
  while ((i > 0) && (path[i-1] != '/'))
    i--;
  path[i] = 0;
  int m = len - (i + 1);
  int l = strlen(baseqname);
  strncpy(path+i, baseqname, m);
  if (l > m)
    return 0;
  if (relqtype)
  {
    if (l >= m)
      return 0;
    path[i+(l++)] = ';';
    if (l >= m)
      return 0;
    strncpy(path+i+l, relqname, m - l);
    l += strlen(relqname);
  }
  return (l > m) ? 0 : m - l;
}

/* special reformatting of ptr address - reverse the name:
"2.1.168.192.in-addr.arpa\0" becomes "192.168.1.2\0"
"b.a.9.8.7.6.5.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.8.b.d.0.1.0.0.2.ip6.arpa\0" becomes "2.0.0.1.0.d.b.8.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.5.6.7.8.9.a.b"
Ouch! It would be nice to add automagic compaction of
IPv6 so it becomes "2001:db8::567:89ab" if that's doable
in a standardized way... */
static char *try_reverse_in_arpa(char *query)
{
  int k, c;
  char new_query[74] = "";
  int l = strlen(query);
  if (l > 12 && l < 29 && !strncmp((char *) query + (l - 13), ".in-addr.arpa", 13)) // IPv4
    k = c = l - 13;
  else if (l > 8 && l < 82 && !strncmp((char *) query + (l - 9), ".ip6.arpa", 9)) // IPv6
    k = c = l - 9;
  else
    return (char *)strdup(query);
  for (; k >= 0; k--)
  {
    if (k == 0 || query[k - 1] == '.')
    {
      strncat ((char *) new_query, (char *) query + k, c - k);
      if (k)
        strcat ((char *) new_query, ".");
      c = --k;
    }
  }
  return (char *)strdup(new_query);
}

// Get file name for query, reversing reverse, disposing disposable
static int get_file_name(char *buf, int len, char *type, char *name, int flags)
{
  if (!(type && *type))
    return 0;
  name = try_reverse_in_arpa(name);
  int r = get_file_path(buf, len, name, type);
  free(name);
  if (flags & DCL_MEMORY)
  {
    strncpy(rr_path, buf, sizeof(rr_path)-1);
    rr_path[sizeof(rr_path)-1] = 0;
  }
  return r;
}

// randomize record order
static void round_robin(char **rrs)
{
  int n, i;
  char *t;
  for (n = 0; rrs[n]; n++);
  for (i = 0; i < n * 2; i++)
  {
    int e;
    e = (long)(random() % n);
    t = rrs[i % n];
    rrs[i % n] = rrs[e];
    rrs[e] = t;
  }
}

// Find resource records
static int directory_search(char *buf, int len, struct stat *inf, int qtype, char *name, int flags, int wrec, int erec)
{
  int sr = -1;
  char *type = "-";
  got_qtype = 0;
  if (flags & DCL_MEMORY)
  {
    found_mtime = 0;
    found_qtype = 0;
    if (!wrec)
    {
      searched_qtype = qtype;
      if (!erec)
        found_cuts = 0;
    }
  }
  if (!name)
    return 0;
  if (qtype)
  {
    type = qtype_fname(qtype);
    if (!(type && *type))
      return 0;
  }
  while (*name == '.')
    name++;
  int tno = 0;
  char *q = name;
  logmsg(LOG_DEBUG+1, "Search: %s %s", type, name);
  if (flags & DCL_DESCEND)
  { //For some lookups, we descend through the domains:
    while (get_file_name(buf, len, type, q, flags & ~(erec ? DCL_MEMORY : 0)) && (sr = stat(buf, inf)) && *q)
    {
      if ((q = strchr(q, '.')))
        q++;
      else
        q = "";
      logmsg(LOG_DEBUG+2, "Search: %s %s", type, q);
      tno++;
    }
  }
  else if (qtype)
  { // Find resource record file:
    if (get_file_name(buf, len, type, name, flags & ~(erec ? DCL_MEMORY : 0)))
      sr = stat(buf, inf);
  }
  else
  { // Find domain name directory:
    if (get_dir_path(buf, len, name))
      sr = stat(buf, inf);
  }
  if (!sr && (qtype ? S_ISREG(inf->st_mode) : S_ISDIR(inf->st_mode)))
  { // We found what we were looking for:
    if (tno)
    { // Since what we found was not exactly what was queried for, DDNS timeouts should not apply:
      inf->st_uid = 0;
      inf->st_gid = 0;
      found_cuts += tno + 1;
    }
    logmsg(LOG_DEBUG+1, "Found: %s %s", type, name);
    got_qtype = qtype;
    if (flags & DCL_MEMORY)
    { // Remember this result:
      strncpy(found_name, q, sizeof(found_name)-1);
      found_name[sizeof(found_name)-1] = 0;
      found_mtime = inf->st_mtime;
      found_qtype = qtype;
    }
    return strlen(buf);
  }
  // Not found, try replacements and wildcards:
  int r;
  if (!erec)
  { // Some records are (partially) interchangeable:
    switch(qtype)
    {
      case REQ_A6:
        if ((r = directory_search(buf, len, inf, REQ_AAAA, name, flags, wrec+1, erec+1)))
          return r;
        break;
      case REQ_AAAA:
        if ((r = directory_search(buf, len, inf, REQ_A6, name, flags, wrec+1, erec+1)))
          return r;
        break;
    }
  }
  if (!wrec && (*name != '*'))
  { // Look for wildcard records:
    char *n = strchr(name, '.');
    if (n && n != name)
    {
      while (*(n+1) == '.')
        n++;
      n--;
      n = strdup(n);
      *n = '*';
      r = directory_search(buf, len, inf, qtype, n, flags, wrec+1, erec);
      free(n);
      if (r)
      { // Since what we found was not exactly what was queried for, DDNS timeouts should not apply:
        inf->st_uid = 0;
        inf->st_gid = 0;
        found_cuts ++;
        return r;
      }
    }
  }
  // No record found...
  if (flags & DCL_MEMORY)
    rr_path[0] = 0;
  return 0;
}

// Read (if needed) and split a resource record set:
char **read_rrset(int qtype, int rqtype, unsigned char *name, char *file, char *data, int datsz, int datl, int maxrec, int maxbuf)
{
  if ((file && *file) && !(*data && (datl > 0)))
  { // We need to read it:
    int fd = open(file, O_RDONLY);
    if (fd == -1)
    {
      logmsg(LOG_WARNING, "Failed to open RRs: %s %s %s %s",qtype_mname(qtype), rqtype ? qtype_mname(rqtype) : "-", name, file);
      return NULL;    
    }
    datl = read(fd, data, datsz);
    close(fd);
    if (datl >= datsz)
    {
      logmsg(LOG_WARNING, "Too big RRs: %s %s %s %s",qtype_mname(qtype), rqtype ? qtype_mname(rqtype) : "-", name, file);
      return NULL;
    }
  }
  else
  {
    file = "-";
    if (datl > datsz)
    {
      logmsg(LOG_WARNING, "Too big RRs: %s %s %s %s",qtype_mname(qtype), rqtype ? qtype_mname(rqtype) : "-", name, file);
      return NULL;
    }
  }
  if (datl <= 0)
  {
    logmsg(LOG_WARNING, "Empty RRs: %s %s %s %s",qtype_mname(qtype), rqtype ? qtype_mname(rqtype) : "-", name, file);
    return NULL;
  }
  data[datl] = 0;
  int i;
  char **r = string_split(data, '\n', maxrec, maxbuf, MAX_RECORD_LEN);
  if (!(r && r[0]))
  {
    logmsg(LOG_WARNING, "Broken RRs: %s %s %s %s",qtype_mname(qtype), rqtype ? qtype_mname(rqtype) : "-", name, file);
    return NULL;
  }
  for (i = 0; r[i]; i++)
  {
    string_chomp((unsigned char *)r[i]);
    if (qtype == REQ_TXT || qtype == REQ_SPF || qtype == REQ_HINFO)
      string_clean_text((unsigned char *)r[i]);
    else
      string_clean_resource((unsigned char *)r[i]);
    if (!*(r[i]))
    {
      logmsg(LOG_WARNING, "Bad RRs: %s %s %s %s",qtype_mname(qtype), rqtype ? qtype_mname(rqtype) : "-", name, file);
      free(r);
      return NULL;
    }
  }
  if (qtype == REQ_A || qtype == REQ_AAAA || qtype == REQ_A6 || qtype == REQ_NS)
    round_robin(r);
  for (i = 0; r[i]; i++)
    if (rqtype)
      logmsg(LOG_DEBUG, "Found: %s.%s %s -> %s", qtype_mname(qtype), qtype_mname(rqtype), name, r[i]);
    else
      logmsg(LOG_DEBUG, "Found: %s %s -> %s", qtype_mname(qtype), name, r[i]);
  return r;
}

// Find and read resource records
char **directory_lookup(int qtype, unsigned char *name, int flags)
{
  struct stat inf;
  if (!qtype)
    return NULL;
  logmsg(LOG_DEBUG+1, "Lookup: %s %s", qtype_mname(qtype), name);
  char path[MAX_PATH_LEN+1];
  path[sizeof(path)-1] = 0;
  if (!directory_search(path, sizeof(path)-1, &inf, qtype, (char *)name, flags, 0, 0))
    return NULL;
  char rr_buf[MAX_BUF+1];
  int c = -1;
  if (inf.st_uid && inf.st_gid && ddns_timeouts && (ddns_purge || ddns_fallback) &&
     (!ddns_usergroup || (inf.st_uid == ddns_user) || (inf.st_gid == ddns_group)) && !ddns_domain_cmp((char *)name))
  { // This dynamic RR is too old...
    unsigned long int age = abs(time(NULL) - inf.st_mtime);
    if (ddns_purge && (age > ddns_purge))
    { // Way too old
      logmsg(LOG_INFO, "Dynamic RR too old to be served: %s %s", qtype_mname(got_qtype), name);
      if (flags & DCL_MEMORY) // Replace original path with empty as if not found:
        rr_path[0] = 0;
      return NULL;
    }
    if (ddns_fallback && (age > ddns_fallback))
    { // Try to find a fallback record:
      char *dfb = ddns_fallback_record(qtype);
      if (dfb)
      {
        got_qtype = qtype;
        if (flags & DCL_MEMORY)
          found_qtype = qtype;
      }
      else
      { // Some records are (partially) interchangeable:
        switch(qtype)
        {
          case REQ_A6:
            if ((dfb = ddns_fallback_record(REQ_AAAA)))
            {
              got_qtype = REQ_AAAA;
              if (flags & DCL_MEMORY)
                found_qtype = REQ_AAAA;
            }
            break;
          case REQ_AAAA:
            if ((dfb = ddns_fallback_record(REQ_A6)))
            {
              got_qtype = REQ_A6;
              if (flags & DCL_MEMORY)
                found_qtype = REQ_A6;
            }
            break;
        }
      }
      if (dfb)
      {
        logmsg(LOG_INFO, "Dynamic RR replaced with fallback: %s %s %s", qtype_mname(got_qtype), name, dfb);
        strncpy(rr_buf, dfb, sizeof(rr_buf)-1);
        rr_buf[sizeof(rr_buf)-1] = 0;
        c = strlen(dfb);
        if ((flags & DCL_MEMORY) && (c > 0))
        { // Add fallback to RR path:
          int l = sizeof(rr_path) - strlen(rr_path);
          if (l > 0)
            strncat(rr_path, ";FALLBACK", l);
        }
      }
    }
  }
  return read_rrset(got_qtype, 0, name, path, rr_buf, sizeof(rr_buf), c, -1, -1);
}

// Check if domain directory exists
int directory_check(int qtype, unsigned char *name, int flags)
{
  logmsg(LOG_DEBUG+1, "Check: %s %s", qtype_mname(qtype), name);
  struct stat inf;
  char path[MAX_PATH_LEN+1];
  path[sizeof(path-1)] = 0;
  return directory_search(path, sizeof(path)-1, &inf, qtype, (char *)name, flags, 0, 0);
}

// Read resource records related to the last found one
char **related_lookup(int baseqtype, int relqtype, unsigned char *name)
{
  logmsg(LOG_DEBUG+2, "Get related: %d.%d %s %s", baseqtype, relqtype, name, rr_path);
  char path[MAX_PATH_LEN+1];
  path[sizeof(path)-1] = 0;
  if (!get_related_path(path, sizeof(path)-1, rr_path, baseqtype, relqtype))
    return NULL; // No path to related record
  logmsg(LOG_DEBUG+1, "Get related: %d.%d %s %s", baseqtype, relqtype, name, path);
  struct stat inf;
  if (stat(path, &inf))
    return NULL; // Related record not found
  char rr_buf[MAX_BUF+1];
  logmsg(LOG_DEBUG+1, "Got related: %d.%d %s %s", baseqtype, relqtype, name, path);
  return read_rrset(baseqtype, relqtype, name, path, rr_buf, sizeof(rr_buf), 0, -1, -1);
}

// Read cached resource records related to the last found one
//char **cache_lookup(int baseqtype, int relqtype, int fresh, unsigned char *name)
//{
//  logmsg(LOG_DEBUG+2, "Get cached: %d.%d %s %s", baseqtype, relqtype, name, rr_path);
//  char path[MAX_PATH_LEN+1];
//  path[sizeof(path)-1] = 0;
//  if (!(get_cache_path(path, sizeof(path)-1, rr_path) && get_related_path(path, sizeof(path)-1, NULL, rr_path, baseqtype, relqtype)))
//    return NULL; // No path to cached record
//  logmsg(LOG_DEBUG+1, "Get cached: %d.%d %s %s", baseqtype, relqtype, name, path);
//  struct stat inf;
//  if (stat(path, &inf))
//    return NULL; // Cached record not found
//  if (fresh && (inf->st_mtime < found_mtime))
//    return NULL; // Cached record too old
//  char rr_buf[MAX_RECORD_LEN+1];
//  logmsg(LOG_DEBUG+1, "Got cached: %d.%d %s %s", baseqtype, relqtype, name, path);
//  return read_rrset(baseqtype, relqtype, name, path, rr_buf, sizeof(rr_buf), 0);
//}

// Get modification time for last found file
struct tm *get_found_mtime()
{
  return localtime(&found_mtime);
}

// Get modification time
//struct tm *get_mtime(int qtype, unsigned char *name)
//{
//  struct stat inf;
//  logmsg(LOG_DEBUG+1, "Time: %s %s", qtype_name(qtype), name);
//  if (!directory_search(dir_buf, sizeof(dir_buf), &inf, qtype, (char *)name, 0, 0, 0))
//    inf.st_mtime = 0;
//  return localtime(&inf.st_mtime);
//}

// Search for domain and subdomains in directory:
int search_dir_for_hosts_in_domain(char **buf, int *bsz, char **hosts, char **first, char **last, int hfllen, char *path, char *domain, int flags)
{
  DIR *dir = opendir(path);
  if (!dir)
    return 0;
  int found = 0;
  logmsg(LOG_DEBUG+3, "Serch hosts: %s %s", domain, path);
  int dl = domain ? strlen(domain) : 0;
  struct stat inf;
  struct dirent *node;
  char stp[MAX_PATH_LEN+1];
  int i;
  while ((node = readdir(dir)))
  {
    int nl = strlen(node->d_name);
    if ((nl < dl) || ((flags & SDH_EXC_DOMAIN) && (nl == dl)))
      continue;
    if ((nl > dl) && (node->d_name[nl-(dl+1)] != '.'))
      continue;
    if (not_db_name(node->d_name, flags))
      continue;
    if (dl && strncmp(node->d_name+(nl-dl), domain, dl))
      continue;
    if (snprintf(stp, sizeof(stp)-1, "%s/%s", path, node->d_name) >= sizeof(stp)-1)
      continue;
    if (stat(stp, &inf) || !S_ISDIR(inf.st_mode))
      continue;
    found++;
    if (bsz)
    {
      (*buf) = realloc((*buf), (*bsz)+nl+2);
      if ((*bsz) > 0)
        (*buf)[(*bsz)++] = '\n';
      strncpy((*buf)+(*bsz), node->d_name, nl+1);
      (*bsz) += nl;
      (*buf)[(*bsz)] = 0;
    }
    if (hfllen)
    {
      nl -= (dl + 1);
      if (nl < 0)
        nl = 0;
      node->d_name[nl] = 0;
      for (i = 0; i < hfllen; i++)
      {
        if (hosts && hosts[i])
        {
          if (first && (!first[i] || (strncmp(node->d_name, first[i], MAX_RECORD_LEN) < 0)) && (strncmp(node->d_name, hosts[i], MAX_RECORD_LEN) > 0))
          {
            first[i] = realloc(first[i], nl+1);
            strncpy(first[i], node->d_name, nl+1);
          }
          if (last && (!last[i] || (strncmp(node->d_name, last[i], MAX_RECORD_LEN) > 0)) && (strncmp(node->d_name, hosts[i], MAX_RECORD_LEN) < 0))
          {
            last[i] = realloc(last[i], nl+1);
            strncpy(last[i], node->d_name, nl+1);
          }
        }
        else
        {
          if (first && (!first[i] || (strncmp(node->d_name, first[i], MAX_RECORD_LEN) < 0)))
          {
            first[i] = realloc(first[i], nl+1);
            strncpy(first[i], node->d_name, nl+1);
          }
          if (last && (!last[i] || (strncmp(node->d_name, last[i], MAX_RECORD_LEN) > 0)))
          {
            last[i] = realloc(last[i], nl+1);
            strncpy(last[i], node->d_name, nl+1);
          }
        }
      }
    }
  }
  closedir(dir);
  return found;
}

// Search for domain and subdomains in hash directories:
int search_for_hosts_in_domain(char **buf, int *bsz, char **hosts, char **first, char **last, int hfllen, char *domain, int flags)
{
  char **hashes = hex_hashes(hash_type);
  if (!hashes)
    return search_dir_for_hosts_in_domain(buf, bsz, hosts, first, last, hfllen, data_dir, domain, flags);
  char path[MAX_PATH_LEN+1];
  int pl = snprintf(path, sizeof(path)-1, "%s/", data_dir);
  if (pl > sizeof(path)-4)
  {
    free(hashes);
    return 0;
  }
  int found = 0;
  for (int i = 0; hashes[i]; i++)
  {
    strncpy(path+pl, hashes[i], 4);
    found += search_dir_for_hosts_in_domain(buf, bsz, hosts, first, last, hfllen, path, domain, flags);
  }
  free(hashes);
  return found;
}

// Lookup domains and subdomains:
char **names_lookup(int qtype, unsigned char *query)
{
  logmsg(LOG_DEBUG+1, "Get hosts: %s %s", qtype_mname(qtype), query);
  char *domain = (char *)query;
  while (domain && ((*domain == '*') || (*domain == '.')))
    domain ++;
  char *buf = NULL;
  int len = 0;
  if (!search_for_hosts_in_domain(&buf, &len, NULL, NULL, NULL, 0, (char *)domain, 0))
    return NULL;
  char **rrs = len ? read_rrset(qtype, 0, query, NULL, buf, len, len, 0, -1) : NULL;
  free(buf);
  return rrs;
}

// List hosts in domain:
char **list_hosts_in_domain(char *domain, int flags)
{
  while (domain && ((*domain == '*') || (*domain == '.')))
    domain ++;
  char *buf = NULL;
  int len = 0;
  if (!search_for_hosts_in_domain(&buf, &len, NULL, NULL, NULL, 0, domain, flags))
    return NULL;
  char **hosts = string_split(buf, '\n', 0, 0, 0);
  free(buf);
  return hosts;
}
