//<: 'dnsdata.c' v0.013 @ 2012-10-08 00:50:13 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS DNS types and consts */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <time.h>

#include "global.h"
#include "knobs.h"
#include "dnsdata.h"
#include "log.h"
#include "strutil.h"

// TTLs
u4 srvttl_time = THREE_DAYS;
u4 defttl_time = ONE_DAY;
u4 dynttl_time = TEN_MINUTES;

/* Supported query types */
const int qtypes_supported[SUPPORTED_RRS] = {
  REQ_A,
  REQ_AAAA,
  REQ_PTR,
  REQ_NS,
  REQ_DNSKEY,
  REQ_MX,
  REQ_SPF,
  REQ_TXT,
  REQ_SRV,
  REQ_A6,
  REQ_SOA,
  REQ_DS,
  REQ_CNAME,
  REQ_HINFO,
  REQ_RRSIG,
  REQ_NSEC,
};

/* Auto-add query types */
const int qtypes_autoadd[AUTOADD_RRS] = {
  REQ_A,
  REQ_AAAA
};

/* Query type name map */
const qtypename qtypes_names[RECOGNIZED_RRS+1] = {
  {REQ_A,       SREQ_A},
  {REQ_AAAA,    SREQ_AAAA},
  {REQ_A6,      SREQ_A6},
  {REQ_SOA,     SREQ_SOA},
  {REQ_CNAME,   SREQ_CNAME},
  {REQ_RRSIG,   SREQ_RRSIG},
  {REQ_NSEC,    SREQ_NSEC},
  {REQ_NS,      SREQ_NS},
  {REQ_DNSKEY,  SREQ_DNSKEY},
  {REQ_PTR,     SREQ_PTR},
  {REQ_MX,      SREQ_MX},
  {REQ_SPF,     SREQ_SPF},
  {REQ_SRV,     SREQ_SRV},
  {REQ_TXT,     SREQ_TXT},
  {REQ_DS,      SREQ_DS},
  {REQ_HINFO,   SREQ_HINFO},
  {REQ_ANY,     SREQ_ANY},
  {REQ_OPT,     SREQ_OPT},
  {0,           ""},
};

// Get index for query type
int qtype_index(int qtype)
{
  for (int i = 0; i < RECOGNIZED_RRS; i++)
    if (qtypes_names[i].qtype == qtype)
      return i;
  return -1;
}

// Get mnomonic name for query type
char* qtype_mname(int qtype)
{
  for (int i = 0; i < RECOGNIZED_RRS; i++)
    if (qtypes_names[i].qtype == qtype)
      return qtypes_names[i].name;
  return qtypes_names[RECOGNIZED_RRS].name;
}

// Get file name for query type
char* qtype_fname(int qtype)
{
  for (int i = 0; i< RECOGNIZED_RRS; i++)
    if (qtypes_names[i].qtype == qtype)
      return qtypes_names[i].name;
  return qtypes_names[RECOGNIZED_RRS].name;
}

// Get query type for file name
int qname_type(char *name)
{
  for (int i = 0; i < RECOGNIZED_RRS; i++)
    if (!strncasecmp(qtypes_names[i].name, name, 7))
      return qtypes_names[i].qtype;
  return qtypes_names[RECOGNIZED_RRS].qtype;
}

/* Do we support this query type? */
int qtype_supported(int qtype)
{
  for (int i = 0; i < SUPPORTED_RRS; i++)
    if (qtypes_supported[i] == qtype)
      return 1;
  return 0;
}

// Output DNS name
int put_name(u1 *buf, u1 **po, u1 *name, u1 *end)
{
  unsigned int c;
  int i, k, n, offset;
  u1 *where = *po;
  offset = *po - buf;
  k = 0;
  while ((c = name[k]) != 0)
  {
    n = 0;
    while ((c = name[k + n]) != 0 && c != '.')
      n++;
    if (n == 0)
      break;
    if (*po + n > end)
    {
      *po = where;
      return -1;
    }
    *(*po)++ = (unsigned char)(n & ~0xC0);
    for (i = 0; i < n; i++)
      *(*po)++ = name[k++];
    if (name[k] == '.')
      k++;
  }
  *(*po)++ = 0;
  return offset;
}

// Output text
int put_text(u1 *buf, u1 **po, char *text, u1 *end)
{
  int r = 1;
  int cl = 0;
  u1 *lp = *po;
  *(*po)++ = cl;
  while (*text)
  {
    if (*po >= end)
      return -1;
    *(*po)++ = (u1)*text++;
    cl++;
    r++;
    if (cl == 255)
    {
      *lp = cl;
      cl = 0;
      lp = *po;
      *(*po)++ = cl;
      r++;
    }
  }
  *lp = cl;
  return r;
}

// Output binary data from hexadecimal string
int put_hex(u1 *buf, u1 **po, char *data, u1 *end)
{
  unsigned long l;
  char *e;
  char hx[3];
  hx[2] = 0;
  while (*data)
  {
    if (*po >= end)
      return -1;
    hx[0] = *data++;
    if (!*data)
      return -2;
    hx[1] = *data++;
    if (!(((l = strtoul(hx, &e, 16)) < 0x100) && !*e))
      return -2;
    *(*po)++ = l;
  }
  return 0;
}

// Output A (IPv4) record
int put_A(u1 *buf, u1 **po, char *addr, u1 *end)
{ // for A records just store the IPv4
  struct in_addr ta;
  ta.s_addr = 0;
  if (inet_pton(AF_INET, addr, &ta) < 1)
    return -2;
  u1 *a = (u1 *)&ta.s_addr;
  for (int i = 0; i < 4; i++) // already in network byte order
    *(*po)++ = *a++;
  return 0;
}

// Output AAAA (IPv6) record
int put_AAAA(u1 *buf, u1 **po, u2 type, char **qres, char *addr, u1*end)
{ // for AAAA records just store the IPv6
  struct in6_addr ta;
  int ti;
  char *t1, *t2;
  switch(type)
  {
    case REQ_AAAA: // AAAA record
      t1 = addr;
      break;
    case REQ_A6:   // No AAAA record, A6 substituted
      ti = 0;
      t2 = next_token(addr, &ti, ';');
      t1 = next_token(addr, &ti, ';');
      if (t1 && t2)
      { // No prefixes in AAAA
        if (atol(t2))
  	  return -3;
      }
      else if (t2) // Empty prefix:
        t1 = t2;
      else
        return -2;
      break;
    default: // Why were we called?
      return -4;
  }
  (*qres) = t1;
  memset(&ta, 0, sizeof(ta));
  if (inet_pton(AF_INET6, t1, &ta) < 1)
    return -2;
  u1 *a = (u1 *)&ta.s6_addr;
  for (int i = 0; i < 16; i++)
    *(*po)++ = *a++;
  return 0;
}

// Output A6 (IPv6 with prefix) record
int put_A6(u1 *buf, u1 **po, u2 type, char **qres, char *addr, u1*end)
{ // for A6 records store prefix, IPv6 address (maight be partial) and name (if)
  struct in6_addr ta;
  char *t1, *t2;
  int ti;
  u2 pl;
  switch(type)
  {
    case REQ_A6:
      ti = 0;
      t2  = next_token(addr, &ti, ';');
      t1 = next_token(addr, &ti, ';');
      if (t1 && t2)
      { // We've got a prefix spec:
        pl = atol(t2);
  	t2 = pl ? next_token(addr, &ti, ';') : NULL;
      }
      else if (t2)
      { // Empty prefix:
        pl = 0;
  	t1 = t2;
  	t2 = NULL;
      }
      else
  	return -2;
      break;
    case REQ_AAAA: // No A6 record, AAAA substituted
      pl = 0;
      t1 = addr;
      t2 = NULL;
      break;
    default: // Why were we called?
      return -4;
  }
  (*qres) = t1;
  memset(&ta, 0, sizeof(ta));
  if (inet_pton(AF_INET6, t1, &ta) < 1)
    return -2;
  *(*po)++ = pl;
  pl = 16 - (pl / 8);
  u1 *a = (u1 *)&ta.s6_addr;
  for (int i = 0; i < pl; i++)
    *(*po)++ = *a++;
  if (!t2)
    return 0;
  int r = put_name(buf, po, (u1 *)t2, end);
  (*qres) = t2;
  return r;
}

// Output SOA (start of Authority) record
int put_SOA(u1 *buf, u1 **po, char **qres, char *addrs, struct tm *tm, u1*end)
{ // for soa records servers we suffix the entry with:
  int i, ti = 0;
  int r1, r2;
  char *t1 = next_token(addrs, &ti, 0);
  char *t2 = next_token(addrs, &ti, 0);
  if (t1 && t2)
  { // we've got a mail address
    for (i = 0; t2[i]; i++)
      if (t2[i] == '@')
        t2[i] = '.';
    r1 = put_name(buf, po, (u1 *)t1, end);
    r2 = put_name(buf, po, (u1 *)t2, end);
  }
  else if (t1)
  {  // We make a mail address
    r1 = put_name(buf, po, (u1 *)t1, end);
    i = strlen(t1)+13;
    t2 = malloc(i);
    snprintf(t2, i, "hostmaster.%s", t1);
    r2 = put_name(buf, po, (u1 *)t2, end);
    free(t2);
    t2 = NULL;
  }
  else
    return -2;
  if ((r1 < 1) || (r2 < 1))
    return min(min(r1,r2),-1);
  (*qres) = t1;
  // Synthesize a serial number:
  u4 sn = ((u4)tm->tm_year + 1900) * 1000000 + ((u4)((u4)tm->tm_yday * 24 * 3600 + tm->tm_hour * 3600 + tm->tm_min * 60 + tm->tm_sec) >> 5);
  // Get a minimum TTL value. Should be same as for NSEC.
  u4 ttl = TTL_POLICY(0, t1);
  *(*po)++ = (sn >> 24) & 0xFF;
  *(*po)++ = (sn >> 16) & 0xFF;
  *(*po)++ = (sn >> 8) & 0xFF;
  *(*po)++ = (sn >> 0) & 0xFF;
  *(*po)++ = (REFRESH >> 24) & 0xFF;
  *(*po)++ = (REFRESH >> 16) & 0xFF;
  *(*po)++ = (REFRESH >> 8) & 0xFF;
  *(*po)++ = (REFRESH >> 0) & 0xFF;
  *(*po)++ = (RETRY >> 24) & 0xFF;
  *(*po)++ = (RETRY >> 16) & 0xFF;
  *(*po)++ = (RETRY >> 8) & 0xFF;
  *(*po)++ = (RETRY >> 0) & 0xFF;
  *(*po)++ = (EXPIRE >> 24) & 0xFF;
  *(*po)++ = (EXPIRE >> 16) & 0xFF;
  *(*po)++ = (EXPIRE >> 8) & 0xFF;
  *(*po)++ = (EXPIRE >> 0) & 0xFF;
  *(*po)++ = (ttl >> 24) & 0xFF;
  *(*po)++ = (ttl >> 16) & 0xFF;
  *(*po)++ = (ttl >> 8) & 0xFF;
  *(*po)++ = (ttl >> 0) & 0xFF;
  return r1;
}

// Output MX (mail exchange) record, with counter based priority
int put_MX(u1 *buf, u1 **po, char *addr, u2 count, u1*end)
{ // for mail servers, prefix with priority number
  *(*po)++ = (10 + count * 10) >> 8;
  *(*po)++ = (10 + count * 10) & 0xFF;
  return put_name(buf, po, (u1 *)addr, end);
}

// Output SRV (service) record, with counter based priority
int put_SRV(u1 *buf, u1 **po, char *addr, u2 count, u1*end)
{ // for services, prefix with priority and weight (0) numbers
  *(*po)++ = (10 + count * 10) >> 8;
  *(*po)++ = (10 + count * 10) & 0xFF;
  *(*po)++ = 0;
  *(*po)++ = 0;
  int ti = 0;
  char *t1 = next_token(addr, &ti, 0);
  char *t2 = next_token(addr, &ti, 0);
  if (!t1 && !t2)
    return -2;
  // Port
  u2 p = atol(t2);
  *(*po)++ = p >> 8;
  *(*po)++ = p & 0xFF;
  return put_name(buf, po, (u1 *)addr, end);
}

// Output HINFO (host info) record
int put_HINFO(u1 *buf, u1 **po, char *info, u1*end)
{ // Two text fields
  int ti = 0;
  char *t1 = next_token(info, &ti, 0);
  char *t2 = next_token(info, &ti, 0);
  int r1 = put_text(buf, po, t1, end);
  int r2 = put_text(buf, po, t2, end);
  return min(min(r1,r2),0);
}

// Output OPT (EDNS pseudo) record
int put_OPT(u1 *buf, u1 **po, u1 optrcode, u2 optflags, u1 *end)
{
  if (*po + 11 > end)
    return -1;
  *(*po)++ = 0;                  // Empty label
  *(*po)++ = REQ_OPT >> 8;
  *(*po)++ = REQ_OPT & 0xFF;
  *(*po)++ = MAX_MSG >> 8;       // EDNS OPT record class is UDP size
  *(*po)++ = MAX_MSG & 0xFF;
  *(*po)++ = optrcode;           // EDNS TTL[0] = RCODE
  *(*po)++ = 0;                  // EDNS TTL[1] = Version
  *(*po)++ = optflags >> 8;      // EDNS TTL[2-3] = Z
  *(*po)++ = optflags & 0xFF;
  *(*po)++ = 0;                  // No RData in our OPT
  *(*po)++ = 0;
  return 0;
}
