//<: 'log.c' v0.004 @ 2012-09-26 18:33:05 UTC :>//

/* What the DNS logging */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#include <errno.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>

#include "knobs.h"

#if LOG_TIME_CLOCK_GETTIME
#elif LOG_TIME_GETTIMEOFDAY
#include <sys/time.h>
#elif LOG_TIME_TIME
#endif

#include "log.h"
#include "version.h"
#include "global.h"

int loglevel = LOG_NOTICE;

static FILE *logfile = NULL;
static FILE *dbgfile = NULL;

static int syslogging = 0;

#define MAX_LOG_LINE 1024
void logstr(int level, const char *format, ...)
{
  va_list args;
  static char buf[MAX_LOG_LINE];
  char *msg = buf;
  buf[0] = 0;
  if (dbgfile || logfile)
  {
    int pl;
    struct tm *lt = NULL;
    long sf = 0;
    char ff[7] = ".%0#ld";
#if LOG_TIME_CLOCK_GETTIME
    struct timespec ct;
    if (!clock_gettime(CLOCK_REALTIME,&ct))
    {
      lt = gmtime(&ct.tv_sec);
      sf = ct.tv_nsec;
      ff[3] = '9';
    }
#elif LOG_TIME_GETTIMEOFDAY
    struct timeval ct;
    if (!gettimeofday(&ct,NULL))
    {
      lt = gmtime(&ct.tv_sec);
      sf = ct.tv_usec;
      ff[3] = '6';
    }
#elif LOG_TIME_TIME
    time_t ct = time(NULL);
    lt = gmtime(&ct);
#endif
    if (lt)
    {
      pl = strftime(msg,28,"%F %T",lt);
      if (pl > 0)
      {
        msg += pl;
	if (ff[3] != '#')
	{
	  pl = snprintf(msg,16,ff,sf);
	  if (pl > 0)
	    msg += pl;
	}
	if (strncpy(msg," UTC  ",6))
	  msg += 6; 
      }
    }
    pl = snprintf(msg,16,"[%lld]  ",(long long int)getpid());
    if (pl > 0)
      msg += pl; 
  }
  va_start(args, format);
  vsnprintf(msg, (MAX_LOG_LINE - 2) - (msg - buf), format, args);
  va_end(args);
  if (dbgfile)
  {
    fprintf(dbgfile, "%s\n", buf);
    fflush(dbgfile);
  }
  if ((level < LOG_DEBUG) || !dbgfile)
  {
    if (logfile)
    {
      fprintf(logfile, "%s\n", buf);
      fflush(logfile);
    }
    if (syslogging && ((level < LOG_INFO) || !logfile))
      syslog((level < LOG_EMERG) ? LOG_EMERG : (level > LOG_DEBUG) ? LOG_DEBUG : level, "%s", msg);
  }
  if ((level < LOG_WARNING) || (loglevel >= 16) || !(syslogging || logfile || dbgfile))
  {
    FILE *file = (level >= LOG_DEBUG) ? stdout : stderr;
    fprintf(file, "%s\n", msg);
    fflush(file);
  }
}

void logerr(const char *format, ...)
{
  va_list args;
  static char msg[128];
  int no = errno;
  char *err = strerror(no);
  va_start(args, format);
  int ml = vsnprintf(msg, sizeof(msg), format, args);
  va_end(args);
  if (err)
    logstr(LOG_CRIT, "%s: %s", (ml > 0) ? msg : format, err);
  else
    logstr(LOG_CRIT, "%s: %i", (ml > 0) ? msg : format, no);
}

void logopen(char *filename)
{
  logclose();
  logfile = fopen(filename, "a");
  if (logfile)
    logmsg(LOG_DEBUG, "log file opened");
  else
    logstr(LOG_ERROR, "log file (%s) open error",filename);
}

void logclose()
{
  if (logfile)
  {
    logmsg(LOG_DEBUG,"log file closing");
    int r = fclose(logfile);
    logfile = NULL;
    if (r)
      logmsg(LOG_WARNING,"log file close error");
  }
}

void dbglogopen(char *filename)
{
  dbglogclose();
  dbgfile = fopen(filename, "a");
  if (dbgfile)
    logmsg(LOG_DEBUG,"debug log file opened");
  else
    logmsg(LOG_ERROR,"debug log file (%s) open error",filename);
}

void dbglogclose() {
  if (dbgfile)
  {
    logmsg(LOG_DEBUG,"debug log file closing");
    int r = fclose(dbgfile);
    dbgfile = NULL;
    if (r)
      logmsg(LOG_WARNING,"debug log file close error");
  }
}

void syslogopen()
{
  syslogclose();
  openlog(program, LOG_NOWAIT|LOG_PID, LOG_DAEMON);
  syslogging ++;
}

void syslogclose()
{
  if (syslogging)
  {
    closelog();
    syslogging = 0;
  }
}

void closelogs()
{
  logclose();
  dbglogclose();
  syslogclose();
}
