//<: 'sockutil.c' v0.015 @ 2012-10-09 23:27:00 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS IP socket stuff */

#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <string.h>

#include "knobs.h"
#include "global.h"
#include "sockutil.h"
#include "strutil.h"
#include "log.h"

static const int yes = 1;

typedef struct {
  int fd;
  int dgcn;
  int ipv;
  int conn;
  char type[4];
  struct sockaddr_storage from;
  socklen_t fromlen;
} dgram_socket;

typedef struct {
  int fd;
  int dgcn;
  int ipv;
  int conn;
  char type[4];
  struct sockaddr_storage from;
  socklen_t fromlen;
  int port;
  struct sockaddr_storage to;
  socklen_t tolen;
} udp_socket;

typedef struct {
  int fd;
  int dgcn;
  int ipv;
  int conn;
  char type[4];
} stream_socket;

typedef struct {
  int fd;
  int dgcn;
  int ipv;
  int conn;
  char type[4];
  struct sockaddr_storage from;
  socklen_t fromlen;
} data_socket;

#define sock_init(s,t,f)  do { if (s) { \
  memset(s, 0, sizeof(*s));             \
  s->fd = -1;                           \
  s->dgcn = f;                          \
  strncpy(s->type, t, 3); }} while (0)

ip_socket *sock_new_dgram()
{
  dgram_socket *sock = malloc(sizeof(dgram_socket));
  sock_init(sock, "sck", SOCK_DGRAM);
  sock->fromlen = sizeof(sock->from);
  return (ip_socket *)sock;
}

ip_socket *sock_new_stream()
{
  stream_socket *sock = malloc(sizeof(stream_socket));
  sock_init(sock, "sck", SOCK_STREAM);
  return (ip_socket *)sock;
}

ip_socket *sock_new_udp()
{
  udp_socket *sock = malloc(sizeof(udp_socket));
  sock_init(sock, "UDP", SOCK_DGRAM);
  sock->fromlen = sizeof(sock->from);
  return (ip_socket *)sock;
}

ip_socket *sock_new_tcp()
{
  stream_socket *sock = malloc(sizeof(stream_socket));
  sock_init(sock, "TCP", SOCK_STREAM);
  return (ip_socket *)sock;
}

ip_socket *sock_new_data_tcp()
{
  data_socket *sock = malloc(sizeof(data_socket));
  sock_init(sock, "TCP", SOCK_STREAM);
  sock->fromlen = sizeof(sock->from);
  sock->conn = 1;
  return (ip_socket *)sock;
}

ip_socket *sock_new_data_stream()
{
  data_socket *sock = malloc(sizeof(data_socket));
  sock_init(sock, "sck", SOCK_STREAM);
  sock->fromlen = sizeof(sock->from);
  sock->conn = 1;
  return (ip_socket *)sock;
}

static char addrstrbuf[256];

char *address_service_string(struct sockaddr *addr, int addrlen, int service)
{
  char pstr[11];
  int i = service ? 1 : 0;
  if (getnameinfo(addr, addrlen, addrstrbuf+i, sizeof(addrstrbuf)-(i*4), service ? pstr : NULL, service ? sizeof(pstr) : 0, NI_NUMERICHOST|NI_NUMERICSERV))
  {
    addrstrbuf[0] = '-';
    addrstrbuf[1] = 0;
  }
  else if (service)
  {
    addrstrbuf[0] = '[';
    i = strlen(addrstrbuf);
    addrstrbuf[i++] = ']';
    addrstrbuf[i++] = ':';
    pstr[sizeof(pstr)-1] = 0;
    strncpy(addrstrbuf+i, pstr, sizeof(addrstrbuf)-i);
  }
  addrstrbuf[sizeof(addrstrbuf)-1] = 0;
  return addrstrbuf;
}

char *fd_address_service_string(int fd, int service)
{
  struct sockaddr_storage sa;
  socklen_t l = sizeof(sa);
  if ((fd == -1) || getsockname(fd, (struct sockaddr *)&sa, &l) || (l > sizeof(sa)))
  {
    addrstrbuf[0] = 0;
    return addrstrbuf;
  }
  return address_service_string((struct sockaddr *)&sa, l, service);
}

char *address_string(struct sockaddr *addr, int addrlen)
{
  return address_service_string(addr, addrlen, 0);
}

char *sock_address_string(ip_socket *sock)
{
  if (sock->conn)
  {
    switch(sock->dgcn)
    {
      case SOCK_STREAM:
        return address_service_string((struct sockaddr *)&(((data_socket *)sock)->from), ((data_socket *)sock)->fromlen, 0);
        break;
      case SOCK_DGRAM:
        return address_service_string((struct sockaddr *)&(((dgram_socket *)sock)->from), ((dgram_socket *)sock)->fromlen, 0);
        break;
    }
  }
  return fd_address_service_string(sock->fd, 0);
}

char *sock_local_address_string(ip_socket *sock)
{
  if (sock->conn && sock->ipv && (sock->type[0] == 'U'))
    return address_service_string((struct sockaddr *)&(((udp_socket *)sock)->to), ((udp_socket *)sock)->tolen, 1);
  return fd_address_service_string(sock->fd, 1);
}

char *sock_service_string(ip_socket *sock)
{
  return fd_address_service_string(sock->fd, 1);
}

char *my_domain_name()
{
  int r = getdomainname(addrstrbuf, sizeof(addrstrbuf));
  addrstrbuf[sizeof(addrstrbuf)-1] = 0;
  if (r || !*addrstrbuf || (*addrstrbuf == '(') || !strchr(addrstrbuf,'.'))
  {
    struct addrinfo hints, *res = NULL;
    r = gethostname(addrstrbuf, sizeof(addrstrbuf));
    addrstrbuf[sizeof(addrstrbuf)-1] = 0;
    if (!r)
    {
      memset(&hints, 0, sizeof(hints));
      hints.ai_flags = AI_CANONNAME;
      hints.ai_family = AF_UNSPEC;
      if (!getaddrinfo(addrstrbuf, NULL, &hints, &res) && res)
      {
        if (res->ai_canonname)
        {
          strncpy(addrstrbuf, res->ai_canonname, sizeof(addrstrbuf));
          addrstrbuf[sizeof(addrstrbuf)-1] = 0;
        }
        freeaddrinfo(res);
      }
    }
  }
  return addrstrbuf;
}

int socket_option_on(int fd, int level, int option)
{
  int sr = setsockopt(fd, level, option, &yes, sizeof(yes));
  if (sr)
    logmsg(LOG_DEBUG,"Set socket option %d.%d: %d %s", level, option, errno, strerror(errno));
  else
  {
    logmsg(LOG_DEBUG+3,"Set socket option %d.%d: %d", level, option, yes);
    int ov;
    socklen_t oz = sizeof(ov);
    int gr = getsockopt(fd, level, option, &ov, &oz);
    if (gr)
      logmsg(LOG_DEBUG,"Get socket option %d.%d: %d %s", level, option, errno, strerror(errno));
    else
    {
      logmsg(LOG_DEBUG+4,"Got socket option %d.%d: %d (%d)", level, option, ov, oz);
      if (ov != yes)
        logmsg(LOG_DEBUG,"Set socket option %d.%d: %d != %d", level, option, ov, yes);
    }
  }
  return sr;
}

void socks_free(ip_socket **socks, int *count)
{
  if (socks)
  {
    while (*count > 0)
    {
      (*count)--;
      sock_free(socks[*count]);
    }
  }
  *count = 0;
}

int create_listen_sockets(ip_socket **socks, int *count, char *address, char *service, int errs)
{
  struct addrinfo hints, *res, *resr;
  int r;
  int max = *count;
  *count = 0;
  char *addresses = strdup((address && *address) ? address : ";");
  char *addr = addresses;
  logmsg(LOG_DEBUG, "listen: [%s]:%s", address, service);
    
  int miss = 0;
  int addri = 0;
  while ((addr = next_token(addresses, &addri, ';')))
  {
    logmsg(LOG_DEBUG, "listen: [%s]:%s", addr, service);
    /* Ask the system what type of address it thinks we've got. */
    char *qaddr = (!*addr || ((*addr == '*') && !*(addr+1))) ? NULL : addr;
    char *qserv = (!*service || ((*service == '*') && !*(service+1))) ? NULL : service;
    logmsg(LOG_DEBUG+5, "listen: [%s]:%s", qaddr, qserv);
    memset(&hints, 0, sizeof(struct addrinfo));
    hints.ai_flags = AI_PASSIVE;
    hints.ai_family = AF_UNSPEC;
    hints.ai_socktype = 0;
    r = getaddrinfo(qaddr, qserv, &hints, &res);
    if (r != 0)
    {
      logmsg(errs ? LOG_ERROR : LOG_DEBUG, "getaddrinfo() failed for [%s]:%s: %d %s", addr, service, r, gai_strerror(r));
      continue;
    }
    if (!res)
    {
      logmsg(errs ? LOG_ERROR : LOG_DEBUG, "getaddrinfo() returned nothing for [%s]:%s", addr, service);
      continue;
    }
    resr = res;
  
    /* Try binding to the addresses we got back. */
    while (res)
    {
      if (res->ai_family && res->ai_protocol && res->ai_socktype)
      {
        logmsg(LOG_DEBUG+2, "hint: %d.%d (%d) %s", res->ai_family, res->ai_protocol, res->ai_socktype, address_service_string(res->ai_addr, res->ai_addrlen, 1));
        int nok = 1;
        int fd;
        if (*count >= max)
          logmsg(LOG_DEBUG,"socket overflow %d.%d (%d) for [%s]%s", res->ai_family, res->ai_protocol, res->ai_socktype, addr, service);
        else if ((fd = socket(res->ai_family, res->ai_socktype, res->ai_protocol)) == -1)
          logmsg(errs ? LOG_ERROR : LOG_DEBUG, "socket error for [%s]%s, %d.%d (%d): %d %s", addr, service, res->ai_family, res->ai_protocol, res->ai_socktype, errno, strerror(errno));
        else
        {
#ifdef IPV6_V6ONLY
          if (res->ai_family == AF_INET6)
            socket_option_on(fd, IPPROTO_IPV6, IPV6_V6ONLY);
#endif
          if (res->ai_protocol == IPPROTO_UDP)
          {
            switch(res->ai_family)
            {
              case AF_INET:
#ifdef IP_PKTINFO
                socket_option_on(fd, SOL_IP, IP_PKTINFO);
#endif
#ifdef IP_RECVDSTADDR
                socket_option_on(fd, IPPROTO_IP, IP_RECVDSTADDR);
#endif
                break;
              case AF_INET6:
#ifdef IPV6_PKTINFO
                socket_option_on(fd, SOL_IPV6, IPV6_PKTINFO);
#endif
#ifdef IPV6_2292PKTINFO
                socket_option_on(fd, SOL_IPV6, IPV6_2292PKTINFO);
#endif
#ifdef IPV6_RECVDSTADDR
                socket_option_on(fd, IPPROTO_IPV6, IPV6_RECVDSTADDR);
#endif
                break;
            } 
          }
#ifdef SO_REUSEADDR
          if (errs)
            socket_option_on(fd, SOL_SOCKET, SO_REUSEADDR);
#endif
          if (bind(fd, res->ai_addr, res->ai_addrlen))
            logmsg(errs ? LOG_ERROR : LOG_DEBUG, "bind error at %s: %d %s", address_service_string(res->ai_addr, res->ai_addrlen, 1), errno, strerror(errno));
          else if ((res->ai_socktype == SOCK_STREAM) && listen(fd, 50))
            logmsg(errs ? LOG_ERROR : LOG_DEBUG, "listen error at %s: %d %s", address_service_string(res->ai_addr, res->ai_addrlen, 1), errno, strerror(errno));
          else
          {
            ip_socket *sock = NULL;
            if ((res->ai_family == AF_INET) || (res->ai_family == AF_INET6))
            {
              if (res->ai_protocol == IPPROTO_UDP)
              {
                sock = sock_new_udp();
                if (res->ai_family == AF_INET)
                  ((udp_socket *)sock)->port = ((struct sockaddr_in *)res->ai_addr)->sin_port;
                else if (res->ai_family == AF_INET6)
                  ((udp_socket *)sock)->port = ((struct sockaddr_in6 *)res->ai_addr)->sin6_port;
                else
                  ((udp_socket *)sock)->port = atoi(service);
              }
              else if (res->ai_protocol == IPPROTO_TCP)
                sock = sock_new_tcp();
              else if (res->ai_socktype == SOCK_DGRAM)
                sock = sock_new_dgram();
              else if (res->ai_socktype == SOCK_STREAM)
                sock = sock_new_stream();
              if (sock)
                sock->ipv = res->ai_family;
            }
            else if (res->ai_socktype == SOCK_DGRAM)
              sock = sock_new_dgram();
            else if (res->ai_socktype == SOCK_STREAM)
              sock = sock_new_stream();
            if (!sock)
              logmsg(LOG_DEBUG+2,"unsupported address %d.%d (%d) for [%s]%s", res->ai_family, res->ai_protocol, res->ai_socktype, addr, service);
            else
            {
              sock->fd = fd;
              socks[(*count)++] = sock;
              logmsg(errs ? LOG_NOTICE : LOG_DEBUG, "accepting %s packets on %s (%d)", sock->type, sock_service_string(sock), *count);
              nok = 0;
            }
          }
          if (nok)
          {
            close(fd);
            fd = -1;
          }
        }
        miss += nok;
      }
      res = res->ai_next;
    }
  
    freeaddrinfo(resr);
  }

  free(addresses);
  if (!*count)
  {
    logmsg(errs ? LOG_ERROR : LOG_DEBUG, "no address succeeded socket(), bind() and listen() for %s[%s]", address, service);
    return miss ? miss : -2;
  }
  else if (miss)
  {
    logmsg(LOG_WARN,"added %d of %d listening sockets for [%s]:%s", *count, miss, address, service);
    return miss;
  }
  else
  {
    logmsg(LOG_DEBUG+1,"added %d listening sockets for [%s]:%s", *count, address, service);
    return 0;
  }
}

int sock_listen(ip_socket **socks, int *sockcnt, char *address, char *service)
{
  int r = create_listen_sockets(socks, sockcnt, address, service, 1);
  if (*sockcnt)
    return 0;
  else
    return r;
}

int sock_listen_random(ip_socket **socks, int *count, int *port)
{
  int r = -3;
  int sockcnt = 0;
  char service[6];
  for (int i = 0; i < 1024; i++)
  {
    socks_free(socks, &sockcnt);
    *port = 10240 + (random() % 20480);
    snprintf(service, 6, "%u", *port);
    sockcnt = *count;
    if (!(r = create_listen_sockets(socks, &sockcnt, "::1;127.0.0.1", service, 0)))
    {
      *count = sockcnt;
      return r;
    }
  }
  if (sockcnt)
  {
    *count = sockcnt;
    return 0;
  }
  socks_free(socks, &sockcnt);
  *port = 0;
  *count = 0;
  return r;
}

ip_socket *sock_accept(ip_socket *srvsock)
{
  if (srvsock->dgcn != SOCK_STREAM)
    return NULL;
  data_socket *sock = (data_socket *)((srvsock->ipv && (srvsock->type[0] == 'T')) ? sock_new_data_tcp() : sock_new_data_stream());
  sock->fd = accept(srvsock->fd, (struct sockaddr *)&(sock->from), &(sock->fromlen));
  if (sock->fd == -1)
  {
    sock_free(sock);
    return NULL;
  }
  sock->ipv = srvsock->ipv;
#if TCP_KEEPALIVE
  socket_option_on(sock->fd, SOL_SOCKET, SO_KEEPALIVE);
#endif
  return (ip_socket *)sock;
}

int sock_receive_from(ip_socket *srvsock, u1 *buf, int sz)
{
  if (srvsock->dgcn != SOCK_DGRAM)
    return -3;
  int r;
  struct msghdr mhdr;
  struct cmsghdr *cmhdr;
  char cmbuf[256];
  struct iovec iov;
  memset(&mhdr, 0, sizeof(struct msghdr));
  memset(&iov, 0, sizeof(struct iovec));
  mhdr.msg_control = cmbuf;
  mhdr.msg_controllen = sizeof(cmbuf);
  mhdr.msg_iov  = &iov;
  mhdr.msg_iovlen = 1;
  mhdr.msg_name = &(((dgram_socket *)srvsock)->from);
  mhdr.msg_namelen = ((dgram_socket *)srvsock)->fromlen;
  ((dgram_socket *)srvsock)->fromlen = sizeof(((dgram_socket *)srvsock)->from);
  if (srvsock->ipv && (srvsock->type[0] == 'U'))
    ((udp_socket *)srvsock)->tolen = 0;
  iov.iov_base = buf;
  iov.iov_len  = sz;
  r = recvmsg(srvsock->fd, &mhdr, MSG_DONTWAIT|MSG_TRUNC);
  ((dgram_socket *)srvsock)->fromlen = mhdr.msg_namelen;
  if (srvsock->ipv && (srvsock->type[0] == 'U'))
  {
    for (cmhdr = CMSG_FIRSTHDR(&mhdr); cmhdr != NULL; cmhdr = CMSG_NXTHDR(&mhdr,cmhdr))
    {
      logmsg(LOG_DEBUG+6,"Recv ctrl-msg: %d.%d", cmhdr->cmsg_level, cmhdr->cmsg_type);
      union {
        struct in6_pktinfo *pi6;
        struct in_pktinfo *pi4;
        struct in6_addr *ia6;
        struct in_addr *ia4;
      } data;
      switch(srvsock->ipv)
      {
        case AF_INET:
#ifdef IP_PKTINFO
          if ((cmhdr->cmsg_level == SOL_IP) && (cmhdr->cmsg_type == IP_PKTINFO))
          {
            data.pi4 = (struct in_pktinfo *)CMSG_DATA(cmhdr);
            ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_family = AF_INET;
            ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_port = ((udp_socket *)srvsock)->port;
            ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_addr = data.pi4->ipi_addr;
            ((udp_socket *)srvsock)->tolen = sizeof(struct sockaddr_in);
            break;
          }
#endif
#ifdef IP_RECVDSTADDR
          if ((cmhdr->cmsg_level == IPPROTO_IP) && (cmhdr->cmsg_type == IP_RECVDSTADDR))
          {
            data.ia4 = (struct in_addr *)CMSG_DATA(cmhdr);
            ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_family = AF_INET;
            ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_port = ((udp_socket *)srvsock)->port;
            ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_addr = *(data.ia4);
            ((udp_socket *)srvsock)->tolen = sizeof(struct sockaddr_in);
            break;
          }
#endif
          break;
        case AF_INET6:
#ifdef IPV6_PKTINFO
          if ((cmhdr->cmsg_level == SOL_IPV6) && (cmhdr->cmsg_type == IPV6_PKTINFO))
          {
            data.pi6 = (struct in6_pktinfo *)CMSG_DATA(cmhdr);
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_family = AF_INET6;
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_port = ((udp_socket *)srvsock)->port;
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_addr = data.pi6->ipi6_addr;
            ((udp_socket *)srvsock)->tolen = sizeof(struct sockaddr_in6);
            break;
          }
#endif
#ifdef IPV6_2292PKTINFO
          if ((cmhdr->cmsg_level == SOL_IPV6) && (cmhdr->cmsg_type == IPV6_2292PKTINFO))
          {
            data.pi6 = (struct in6_pktinfo *)CMSG_DATA(cmhdr);
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_family = AF_INET6;
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_port = ((udp_socket *)srvsock)->port;
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_addr = data.pi6->ipi6_addr;
            ((udp_socket *)srvsock)->tolen = sizeof(struct sockaddr_in6);
            break;
          }
#endif
#ifdef IPV6_RECVDSTADDR
          if ((srvsock->ipv == AF_INET6) && ((cmhdr->cmsg_level == IPPROTO_IPV6) && (cmhdr->cmsg_type == IPV6_RECVDSTADDR)))
          {
            data.ia6 = (struct in6_addr *)CMSG_DATA(cmhdr);
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_family = AF_INET6;
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_port = srvsock->port;
            ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_addr = *(data.ia6);
            ((udp_socket *)srvsock)->tolen = sizeof(struct sockaddr_in6);
            break;
          }
#endif
          break;
      }
    }
  }
  if ((r > 0) && (r < sz))
    memset(buf + r, 0, sz - r);
  if ((r >= 0) && (mhdr.msg_flags && MSG_TRUNC))
    r = -2;
  return r;
}

int sock_send_to(ip_socket *srvsock, u1 *buf, int sz)
{
  if (srvsock->dgcn != SOCK_DGRAM)
    return -3;
  int r;
  struct msghdr mhdr;
  struct iovec iov;
  int mcz = 0;
  memset(&mhdr, 0, sizeof(struct msghdr));
  memset(&iov, 0, sizeof(struct iovec));
  mhdr.msg_iov  = &iov;
  mhdr.msg_iovlen = 1;
  mhdr.msg_name = &(((dgram_socket *)srvsock)->from);
  mhdr.msg_namelen = ((dgram_socket *)srvsock)->fromlen;
  iov.iov_base = buf;
  iov.iov_len  = sz;
  if (srvsock->ipv && (srvsock->type[0] == 'U') && (((udp_socket *)srvsock)->tolen))
  {
    switch(srvsock->ipv)
    {
      case AF_INET:
#ifdef IP_PKTINFO
        mcz += CMSG_SPACE(sizeof(struct in_pktinfo));
#endif
#ifdef IPV_SENDSRCADDR
        mcz += CMSG_SPACE(sizeof(struct in_addr));
#endif
        break;
      case AF_INET6:
#ifdef IPV6_PKTINFO
        mcz += CMSG_SPACE(sizeof(struct in6_pktinfo));
#endif
#ifdef IPV6_2292PKTINFO
        mcz += CMSG_SPACE(sizeof(struct in6_pktinfo));
#endif
#ifdef IPV6_SENDSRCADDR
        mcz += CMSG_SPACE(sizeof(struct in6_addr));
#endif
        break;
    }
    if (mcz > 0)
    {
      mhdr.msg_control = malloc(mcz);
      mhdr.msg_controllen = mcz;
      memset(mhdr.msg_control, 0, mcz);
      struct cmsghdr *cmhdr = NULL;
      union {
        struct in6_pktinfo *pi6;
        struct in_pktinfo *pi4;
        struct in6_addr *ia6;
        struct in_addr *ia4;
      } data;
      switch(srvsock->ipv)
      {
        case AF_INET:
#ifdef IP_PKTINFO
          cmhdr = cmhdr ? CMSG_NXTHDR(&mhdr,cmhdr) : CMSG_FIRSTHDR(&mhdr);
          cmhdr->cmsg_level = SOL_IP;
	  cmhdr->cmsg_type = IP_PKTINFO;
	  cmhdr->cmsg_len = CMSG_LEN(sizeof(struct in_pktinfo));
	  data.pi4 = (struct in_pktinfo *)CMSG_DATA(cmhdr);
	  memset(data.pi4, 0, sizeof(*(data.pi4)));
	  data.pi4->ipi_spec_dst = ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_addr;
#endif
#ifdef IP_SENDSRCADDR
          cmhdr = cmhdr ? CMSG_NXTHDR(&mhdr,cmhdr) : CMSG_FIRSTHDR(&mhdr);
          cmhdr->cmsg_level = IPPROTO_IP;
	  cmhdr->cmsg_type = IP_SENDSRCADDR;
	  cmhdr->cmsg_len = CMSG_LEN(sizeof(struct in_addr));
	  data.ia4 = (struct in_addr *)CMSG_DATA(cmhdr);
	  *(data.ia4) = ((struct sockaddr_in *)&(((udp_socket *)srvsock)->to))->sin_addr;
#endif
	  break;
        case AF_INET6:
#ifdef IPV6_PKTINFO
          cmhdr = cmhdr ? CMSG_NXTHDR(&mhdr,cmhdr) : CMSG_FIRSTHDR(&mhdr);
          cmhdr->cmsg_level = SOL_IPV6;
	  cmhdr->cmsg_type = IPV6_PKTINFO;
	  cmhdr->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
	  data.pi6 = (struct in6_pktinfo *)CMSG_DATA(cmhdr);
	  memset(data.pi6, 0, sizeof(*(data.pi6)));
	  data.pi6->ipi6_addr = ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_addr;
#endif
#ifdef IPV6_2292PKTINFO
          cmhdr = cmhdr ? CMSG_NXTHDR(&mhdr,cmhdr) : CMSG_FIRSTHDR(&mhdr);
          cmhdr->cmsg_level = SOL_IPV6;
	  cmhdr->cmsg_type = IPV6_2292PKTINFO;
	  cmhdr->cmsg_len = CMSG_LEN(sizeof(struct in6_pktinfo));
	  data.pi6 = (struct in6_pktinfo *)CMSG_DATA(cmhdr);
	  memset(data.pi6, 0, sizeof(*(data.pi6)));
	  data.pi6->ipi6_addr = ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_addr;
#endif
#ifdef IPV6_SENDSRCADDR
          cmhdr = cmhdr ? CMSG_NXTHDR(&mhdr,cmhdr) : CMSG_FIRSTHDR(&mhdr);
          cmhdr->cmsg_level = IPPROTO_IPV6;
	  cmhdr->cmsg_type = IPV6_SENDSRCADDR;
	  cmhdr->cmsg_len = CMSG_LEN(sizeof(struct in6_addr));
	  data.ia4 = (struct in6_addr *)CMSG_DATA(cmhdr);
	  *(data.ia6) = ((struct sockaddr_in6 *)&(((udp_socket *)srvsock)->to))->sin6_addr;
#endif
	  break;
      }
    }
  }
  r = sendmsg(srvsock->fd, &mhdr, 0);
  if (mhdr.msg_control && (mcz > 0))
    free(mhdr.msg_control);
  return r;
}

int sock_send(ip_socket *sock, u1 *buf, int sz, int block)
{
  if (!sock->conn || (sock->dgcn != SOCK_STREAM))
    return -3;
  int f = 0;
  if (!block)
  {
    if ((MAX_IO_BLOCK > 0) && (sz > MAX_IO_BLOCK))
      sz = MAX_IO_BLOCK;
    f |= MSG_DONTWAIT;
  }
  return send(sock->fd, buf, sz, f);
}

int sock_receive(ip_socket *sock, u1 *buf, int sz, int block)
{
  if (!sock->conn || (sock->dgcn != SOCK_STREAM))
    return -3;
  int f = 0;
  if (!block)
    f |= MSG_DONTWAIT;
  return recv(sock->fd, buf, sz, f);
}

