//<: 'wtdnsd.c' v0.075 @ 2012-10-16 15:37:58 UTC (Jonas@Ubbe-III) "Jonas Eckerman" <jonas@truls.org> :>//

/* What the DNS daemon */

#include <limits.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <string.h>
#include <assert.h>
#include <time.h>
#include <signal.h>
#include <errno.h>

#include "knobs.h"
#include "global.h"
#include "sockutil.h"
#include "dnsdata.h"
#include "dnssec.h"
#include "strutil.h"
#include "dir.h"
#include "log.h"
#include "version.h"

#if MAKEDBGD && !MAKEMEMW
#include <mcheck.h>
#endif

#define OPM_SEND        'S'
#define OPM_RECV        'R'
#define OPM_WAIT        'W'

enum {
    ANSWERS = 0,
    SERVERS = 1,
    EXTRAS = 2,
    ANS_BLOCKS = 3
};

static long long int idcounter = 0;

struct interface {
    struct interface *next;
    ip_socket *sock;
    int ctl;
};

static struct interface interfaces = { NULL, NULL, 0 };

struct connection {
    struct connection *next;
    u1 *buf, *bufi;
    long long int cid;
    int ctl;
    char opmode;
    u1 rcerr;
    time_t last_activity, conn_at, ioi_at;
    u2 len, alloced;
    ip_socket *sock;
};

static int nconnections = 0;
static struct connection connections = { NULL, NULL, NULL, -1, 0, 0, 0, 0, 0, 0, 0, 0 };

// Control token domain:
char *control_domain = NULL;

#if COUNTER_LOG || MACROS
#define COUNTERS 1
#endif

#if COUNTERS
// Statistics counter structure:
struct {
  unsigned long long int conn_tot, conn_tcp, conn_udp, conn_ip4, conn_ip6;
  unsigned long long int msgs_tot, msgs_tcp, msgs_udp, msgs_opt, msgs_bad, msgs_sec;
  unsigned long long int rsps_tot, rsps_tcp, rsps_udp, rsps_opt, rsps_non, rsps_err, rsps_big;
  unsigned long long int szin_tot, szin_tcp, szin_udp, szin_opt, szok_opt;
  unsigned long long int szut_tot, szut_tcp, szut_udp, szut_opt;
  unsigned long long int msgs_typ[RECOGNIZED_RRS+1];
  unsigned long long int rsps_typ[RECOGNIZED_RRS+1];
} counters;

// counter incrementation macros:
#define cinc_ipx(s)     if (sock_ipv4(s)) counters.conn_ip4++; else if (sock_ipv6(s)) counters.conn_ip6++;
#define cinc_udp()      do { if (!iface->ctl) { counters.conn_tot++; counters.conn_udp++; cinc_ipx(iface->sock) }} while (0)
#define cinc_tcp()      do { if (!iface->ctl) { counters.conn_tot++; counters.conn_tcp++; cinc_ipx(iface->sock) }} while (0)
#define cinc_sec()      do { if (!conn->ctl) counters.msgs_sec++; } while (0)
#define cinc_bad()      do { if (!conn->ctl) counters.msgs_bad++; } while (0)
#define cinc_big()      do { if (!conn->ctl) counters.rsps_big++; } while (0)
#define cinc_err()      do { if (!conn->ctl) counters.rsps_err++; } while (0)
#define cinc_non()      do { if (!conn->ctl) counters.rsps_non++; } while (0)
#define cinc_tyr(t)     do { if (!conn->ctl) { counters.rsps_typ[qtype_index(t)+1]++; }} while (0)
#define cinc_tym(c,t)   do { if (!conn->ctl) { if ((c == CLASS_ANY) || (c == CLASS_INET)) counters.msgs_typ[qtype_index(t)+1]++; }} while (0)
#define cinc_opt(z,m)   do { if (!conn->ctl) { counters.msgs_opt++; counters.szin_opt+=z; counters.szok_opt+=m; }} while (0)
#define cinc_msg(z)     do { if (!conn->ctl) { counters.msgs_tot++; counters.szin_tot+=z; if (conn->sock->ipv) {        \
                             if (conn->sock->type[0] == 'T') {                                                          \
                               counters.msgs_tcp++;                                                                     \
                               counters.szin_tcp+=z;                                                                    \
                             } else if (conn->sock->type[0] == 'U') {                                                   \
                               counters.msgs_udp++;                                                                     \
                               counters.szin_udp+=z;                                                                    \
                             }}}} while (0)
#define cinc_res(z,o) do { if (!conn->ctl) { counters.rsps_tot++; counters.szut_tot+=z; \
                             if (o) {                                                   \
                               counters.rsps_opt++;                                     \
                               counters.szut_opt+=z;}                                   \
                             if (conn->sock->ipv) {                                     \
                               if (conn->sock->type[0] == 'T') {                        \
                                 counters.rsps_tcp++;                                   \
                                 counters.szut_tcp+=z;                                  \
                               } else if (conn->sock->type[0] == 'U') {                 \
                                 counters.rsps_udp++;                                   \
                                 counters.szut_udp+=z;                                  \
                             }}}} while (0)
#else
#define cinc_udp()      do { } while (0)
#define cinc_tcp()      do { } while (0)
#define cinc_sec()      do { } while (0)
#define cinc_bad()      do { } while (0)
#define cinc_big()      do { } while (0)
#define cinc_err()      do { } while (0)
#define cinc_tym(c,t)   do { } while (0)
#define cinc_tyr(t)     do { } while (0)
#define cinc_opt(z,m)   do { } while (0)
#define cinc_msg(z,h)   do { } while (0)
#define cinc_res(z,h,o) do { } while (0)
#endif

#define SS_CONN 'C'
#define SS_MSGS 'M'
#define SS_RSPS 'R'
#define SS_SIZE 'Z'
#define SS_TIME 'T'
#define SS_VERS 'V'
#define SS_MSGT 'm'
#define SS_RSPT 'r'
time_t start_time;

static int get_name(u1 *msg, u1 **here, u1 *end, unsigned char *string, int string_len, int ctlq, int k)
{
  unsigned int len;
  int i;
  if (*here >= end)
    return -1;
  while ((len = **here) != 0)
  {
    *here += 1;
    if ((len & 0xc0) == 0xc0)
    {
      unsigned int offset;
      u1 *p;
      if (*here >= end)
        return -1;
      offset = ((len & ~0xc0) << 8) + **here;
      p = &msg[offset];
      k = get_name(msg, &p, end, string, string_len, ctlq, k);
      if (k == -1)
        return k;
      break;
    }
    else if ((len & 0xc0))
    {
      return -1;
    }
    else
    {
      for (i = 0; i < len; i++)
      {
        if (*here >= end)
	  return -1;
	if (k < string_len - 1)
	{
	  if (bad_name_char(**here) && !(ctlq && !k && (**here == '*')))
	  {
	    return -1;
	  }
	  if (string != NULL)
	    string[k++] = **here;
	}
	*here += 1;
      }
      if ((k < string_len - 1) && (string != NULL))
        string[k++] = '.';
    }
  }
  *here += 1;
  if (string != NULL)
    string[k] = '\0';
  return k;
}

int write_file(char *filename, char *text, mode_t mode)
{
  int tz = strlen(text);
  int fd = open(filename, O_WRONLY | O_CREAT | O_TRUNC, mode);
  if (fd == -1)
    return 0;
  int rw = write(fd, text, tz);
  int rc = close(fd);
  return (!rc && (rw == tz));
}

void write_pid(char *filename)
{
  char buf[20];
  snprintf(buf, sizeof(buf), "%lld\n", (long long int)getpid());
  if (write_file(filename, buf, NEW_FILE_MODE))
    logmsg(LOG_DEBUG, "pid file created: %s", filename);
  else
    logmsg(LOG_DEBUG, "failed to create pid file: %s", filename);
}

// Format counters
char *counter_string(char cntr)
{
  static char buf[256];
#if COUNTERS
  char *s;
  int i,l,r;
  long long int t;
#endif
  switch(cntr)
  {
    case SS_VERS:
      snprintf(buf, sizeof(buf), "%s (%s)", versionstring, MAKEDATE);
      break;
    case SS_TIME:
      format_time_since(buf, sizeof(buf), start_time);
      break;
#if COUNTERS
    case SS_CONN:
      snprintf(buf, sizeof(buf), "cTot:%llu, cUDP:%llu, cTCP:%llu, cIPv4:%llu, cIPv6:%llu",
                counters.conn_tot, counters.conn_udp, counters.conn_tcp, counters.conn_ip4, counters.conn_ip6);
      break;
    case SS_MSGS:
      snprintf(buf, sizeof(buf), "mTot:%llu, mUDP:%llu, mTCP:%llu, mOPT:%llu, mDO:%llu, mBad:%llu",
                counters.msgs_tot, counters.msgs_udp, counters.msgs_tcp, counters.msgs_opt, counters.msgs_sec, counters.msgs_bad);
      break;
    case SS_RSPS:
      snprintf(buf, sizeof(buf), "rTot:%llu, rUDP:%llu, rTCP:%llu, rOPT:%llu, rNon:%llu, rErr:%llu, rBig:%llu",
                counters.rsps_tot, counters.rsps_udp, counters.rsps_tcp, counters.rsps_opt,
                counters.rsps_non, counters.rsps_err, counters.rsps_big);
      break;
    case SS_MSGT:
      s = buf;
      l = sizeof(buf);
      t = 0;
      for (i = 0; (i <= RECOGNIZED_RRS) && (l > 0); i++)
        if (counters.msgs_typ[i] > 0)
        {
          r = snprintf(s, l, "tm%s:%llu, ",i ? qtypes_names[i-1].name : "Other", counters.msgs_typ[i]);
          l -= r;
          s += r;
          t += counters.msgs_typ[i];
        }
      if (l > 0)
        snprintf(s, l, "tmTot:%llu", t);
      break;
    case SS_RSPT:
      s = buf;
      l = sizeof(buf);
      t = 0;
      for (i = 0; (i <= RECOGNIZED_RRS) && (l > 0); i++)
        if (counters.rsps_typ[i] > 0)
        {
          r = snprintf(s, l, "tr%s:%llu, ",i ? qtypes_names[i-1].name : "Other", counters.rsps_typ[i]);
          l -= r;
          s += r;
          t += counters.rsps_typ[i];
        }
      if (l > 0)
        snprintf(s, l, "trTot:%llu", t);
      break;
    case SS_SIZE:
      snprintf(buf, sizeof(buf), "ziMsg:%0.0f, ziUDP:%0.0f, ziTCP:%0.0f, ziOPT:%0.0f, zmOPT:%0.0f, zoMsg:%0.0f, zoUDP:%0.0f, zoTCP:%0.0f, zoOPT:%0.0f",
                (counters.szin_tot && counters.msgs_tot) ? (double)counters.szin_tot/counters.msgs_tot : 0.0,
                (counters.szin_udp && counters.msgs_udp) ? (double)counters.szin_udp/counters.msgs_udp : 0.0,
                (counters.szin_tcp && counters.msgs_tcp) ? (double)counters.szin_tcp/counters.msgs_tcp : 0.0,
                (counters.szin_opt && counters.msgs_opt) ? (double)counters.szin_opt/counters.msgs_opt : 0.0,
                (counters.szok_opt && counters.msgs_opt) ? (double)counters.szok_opt/counters.msgs_opt : 0.0,
                (counters.szut_tot && counters.rsps_tot) ? (double)counters.szut_tot/counters.rsps_tot : 0.0,
                (counters.szut_udp && counters.rsps_udp) ? (double)counters.szut_udp/counters.rsps_udp : 0.0,
                (counters.szut_tcp && counters.rsps_tcp) ? (double)counters.szut_tcp/counters.rsps_tcp : 0.0,
                (counters.szut_opt && counters.rsps_opt) ? (double)counters.szut_opt/counters.rsps_opt : 0.0);
      break;
#endif
    default:
      buf[0] = 0;
  }
  buf[sizeof(buf)-1] = 0;
  return buf;
}

#if COUNTERS
// Log current counters
#define log_counter(s,c) do { char *l = counter_string(c); if (l && *l) logmsg(LOG_INFO,"%s: %s",s,l); } while (0);
static void log_counters()
{
  log_counter("conn",SS_TIME);
  log_counter("conn",SS_CONN);
  log_counter("msgs",SS_MSGS);
  log_counter("rsps",SS_RSPS);
  log_counter("msgt",SS_MSGT);
  log_counter("rspt",SS_RSPT);
  log_counter("size",SS_SIZE);
}
#endif

#if MACROS
// Expand macro strings (%%wtdns:macro%%)
char *expand_macro(char *txt)
{
  if ((*txt++ != '%') || (*txt++) != '%')
    return NULL;
  char *e = txt + strlen(txt);
  if ((e == txt) || (*--e != '%') || (*--e != '%'))
    return NULL;
  if (strncmp(txt,"wtdns:",6))
    return NULL;
  txt += 6;
  char *mac = strndup(txt, e-txt);
  char mc = 0;
  if (!strcmp(mac, "connections"))
    mc = SS_CONN;
  else if (!strcmp(mac, "queries"))
    mc = SS_MSGS;
  else if (!strcmp(mac, "responses"))
    mc = SS_RSPS;
  else if (!strcmp(mac, "querytypes"))
    mc = SS_MSGS;
  else if (!strcmp(mac, "responstypes"))
    mc = SS_RSPS;
  else if (!strcmp(mac, "sizes"))
    mc = SS_SIZE;
  else if (!strcmp(mac, "time"))
    mc = SS_TIME;
  else if (!strcmp(mac, "version"))
    mc = SS_VERS;
  else
  {
    for (e = mac; *e; e++)
      if (!(((*e >= 'a') && (*e <= 'z')) || ((*e >= 'A') && (*e <= 'Z')) || ((*e >= '0') && (*e <= '9'))))
      {
        free(mac);
        return NULL;
      }
    logmsg(LOG_WARNING,"Unknown macro: '%s'",mac);
  }
  free(mac);
  return counter_string(mc);
}
#endif

void finish(int ec)
{
  struct interface *i;
  for (i = interfaces.next; i;)
  {
    struct interface *t = i->next;
    sock_free(i->sock);
    free(i);
    i = t;
  }
  if (control_domain)
    free(control_domain);
  closelogs();
  free_globals();
#if MAKEDBGD && !MAKEMEMW
  muntrace();
#endif
  exit(ec);
}

#define killtcp(c)						  \
	do {							  \
	    struct connection *tmp;				  \
	    sock_free((c)->sock);				  \
	    nconnections--;                                       \
	    tmp = (c)->next, free (c->buf), free (c), (c) = tmp;  \
	} while (0)

void interrupt(int signum)
{
  if (signum)
    logmsg(LOG_INFO,"interrupted (%i)", signum);
  struct connection *c;
  for (c = &connections; c->next;)
  {
    killtcp (c->next);
    c = c->next;
  }
#if COUNTER_LOG
  log_counters();
#endif
  logmsg(signum ? LOG_NOTICE : LOG_WARNING,"exiting");
  finish(0);
}

char *constr(struct connection *c, int local)
{
  static char cstr[96];
  int l = snprintf(cstr, sizeof(cstr), "#%llu %s %s", c->cid, c->sock->type, sock_address_string(c->sock));
  if (local && c->sock->conn && (l < sizeof(cstr)))
    snprintf(cstr+l, sizeof(cstr)-l, "->%s", sock_local_address_string(c->sock));
  cstr[sizeof(cstr)-1] = 0;
  return cstr;
}

typedef struct qsecdata {
  struct qsecdata *next;
  u2 type;
  u2 reltype;
  u1 section;
  u1 *query;
  int nameptr;
  int related;
  int prio;
  struct qsecdata *same;
} query_section;

query_section *add_qsec(query_section *list, query_section *here, u1 section, int prio, u2 qtype, u2 reltype, int related, int nameptr, u1 *query)
{
  query_section *qst, *same = NULL, *sect = NULL, *prev = NULL, *last = NULL, *fnot = NULL;
  for (qst = list; qst; qst = qst->next)
  {
    last = qst;
    if ((qst->section == section) && (qst->prio <= prio))
      sect = qst;
    if ((qst->section < section) || ((qst->section == section) && (qst->prio <= prio)))
      prev = qst;
    if (!same && !strncmp((char *)(qst->query), (char *)query, MAX_RECORD_LEN))
      same = qst;
    if ((qst->type == qtype) && (qst->reltype == reltype) && (qst->related == related) &&
        (!strncmp((char *)qst->query, (char *)query, MAX_RECORD_LEN)))
    {
      if (qst->section <= section)
        return NULL;
      else if (!fnot && (qst->section < ANS_BLOCKS))
        fnot = qst;
    }
  }
  for (qst = fnot; qst; qst = qst->next)
    if ((qst->section < ANS_BLOCKS) && (qst->type == qtype) && (qst->reltype == reltype) && (qst->related == related) &&
        (!strncmp((char *)qst->query, (char *)query, MAX_RECORD_LEN)))
      qst->section = ANS_BLOCKS;
  qst = malloc(sizeof(query_section));
  qst->type = qtype;
  qst->reltype = reltype;
  qst->related = related;
  qst->section = section;
  qst->nameptr = nameptr;
  qst->prio = prio;
  qst->query = same ? same->query : (u1 *)strdup((char *)query);
  qst->same = same;
  if (!here)
    here = sect ? sect : prev ? prev : last;
  if (here) {
    qst->next = here->next;
    here->next = qst;
  } else {            
    qst->next = NULL;
  }
  return qst;
}

void add_qsec_extra(query_section *list, int nameptr, u1 *query)
{
  for (int k = 0; k < AUTOADD_RRS; k ++)
    if (add_qsec(list, NULL, EXTRAS, 0, qtypes_autoadd[k], 0, 0, nameptr, query))
      logmsg(LOG_DEBUG+2, "Extra: %u %s", qtypes_autoadd[k], query);
}

void add_qsec_NSEC(query_section *list, u1 *query, int flags)
{
  char *psec1 = NULL, *psec2 = NULL;
  if (!find_nsecs(&psec1, &psec2, query, flags))
    logmsg(LOG_WARN,"No NSEC host found for 's'", query);
  else
  {
    if (psec1)
    {
      if (*psec1)
        add_qsec(list, NULL, SERVERS, -5, REQ_NSEC, 0, 0, -1, (u1 *)psec1);
      free(psec1);
    }
    if (psec2)
    {
      if (*psec2)
        add_qsec(list, NULL, SERVERS, -5, REQ_NSEC, 0, 0, -1, (u1 *)psec2);
      free(psec2);
    }
  }
}

#define add_result(s) do { if (s && *s) {  \
  tw1 = strlen(s);                         \
  ts2 = realloc(ts2, tl1 + tw1 + 2);       \
  if (tl1) ts2[tl1++] = '\n';              \
  strncpy(ts2+tl1, s, tw1+1);              \
  tl1 += tw1; }} while (0)

char **control_lookup(u2 qtype)
{ // Control channel request
  char *ts1, *ts2 = NULL;
  u4 tl1 = 0;
  u2 tw1;
  struct interface *ifp;
  switch(qtype)
  {
    case REQ_A:         // List IPv4 listen addresses
      for (ifp = interfaces.next; ifp; ifp = ifp->next)
        if (!ifp->ctl && sock_ipv4(ifp->sock))
          add_result(sock_address_string(ifp->sock));
      break;
    case REQ_A6:
    case REQ_AAAA:      // List IPv6 listen addresses
      for (ifp = interfaces.next; ifp; ifp = ifp->next)
        if (!ifp->ctl && sock_ipv6(ifp->sock))
          add_result(sock_address_string(ifp->sock));
      break;
    case REQ_SRV:       // List listen addresses and ports
      for (ifp = interfaces.next; ifp; ifp = ifp->next)
        if (!ifp->ctl)
          add_result(sock_service_string(ifp->sock));
      break;
    case REQ_TXT:       // List counters etc
      add_result(counter_string(SS_VERS));
      add_result(counter_string(SS_TIME));
#if COUNTERS
      add_result(counter_string(SS_CONN));
      add_result(counter_string(SS_MSGS));
      add_result(counter_string(SS_RSPS));
      add_result(counter_string(SS_MSGT));
      add_result(counter_string(SS_RSPT));
      add_result(counter_string(SS_SIZE));
#endif
      break;
    case REQ_PTR:       // Show host fqdn
      add_result(my_domain_name());
      break;
    case REQ_SOA:       // Show dynamic dns SOA
      return directory_lookup(qtype, (u1 *)ddns_domain, DCL_DESCEND|DCL_MEMORY);
      break;
    case REQ_NS:        // Show dynamic dns NS
      return directory_lookup(qtype, (u1 *)ddns_domain, DCL_DESCEND|DCL_MEMORY);
      break;
    case REQ_HINFO:    // Show uname data
      ts1 = host_info();
      if (ts1)
      {
        add_result(ts1);
        free(ts1);
      }
      break;
  }
  if (!ts2)
    return NULL;
  char **lr = string_split(ts2, '\n', -1, MAX_BUF, MAX_RECORD_LEN);
  free(ts2);
  return lr;
}

#define RSP_NON  0
#define RSP_UDP  1
#define RSP_TCP  2
#define RSP_ALL  (RSP_UDP | RSP_TCP)

#define packet_error(e,a,...) do {                                                                                                        \
  cinc_bad();                                                                                                                             \
  logmsg(LOG_NOTICE,__VA_ARGS__);                                                                                                         \
  if (!((a & RSP_ALL) || ((conn->sock->type[0] == 'U') || (a & RSP_UDP)) || ((conn->sock->type[0] == 'T') && (a & RSP_TCP)))) return -1;  \
  flags |= e;                                                                                                                             \
  goto empty_packet; } while (0)

#define check_size(z,a,w1,w2) do {                                                                                                         \
  if (pi1 + z > in_end) {                                                                                                                  \
    cinc_bad();                                                                                                                            \
    logmsg(LOG_NOTICE, "(%s) Broken package: %u.%u %u > %u", constr(conn,0), w1, w2, z, in_end - pi1);                                     \
    if (!((a & RSP_ALL) || ((conn->sock->type[0] == 'U') || (a & RSP_UDP)) || ((conn->sock->type[0] == 'T') && (a & RSP_TCP)))) return -1; \
    flags |= RC_FORMERR;                                                                                                                   \
    goto empty_packet; }} while (0)

#define format_error(s) do {                                    \
  logmsg(LOG_WARN, "Bad format: %s %s", qtype_mname(qtype), s);  \
  flags = (flags & ~RCODE) | RC_SERVFAIL;                       \
  skip_this ++;                                                 \
  goto skip_record; } while (0)

static int process_packet(struct connection *conn, u1 *in_buf, int packet_len, u1 **send_buf, u2 *send_len)
{
  logmsg(LOG_DEBUG, "(%s) received packet: %i", constr(conn,1), packet_len);
  u1 _out_buf[MAX_BUF + 2];
  u1 *out_buf = &_out_buf[2];
  u1 query[MAX_RECORD_LEN + 1];
  query_section *qsecs = NULL, *qsec;
  int query_len, name_pointer, i, j, k;
  u2 id, qflags, querycnt, nqueries = 0, nanswers = 0, nservers = 0, nextras = 0, qtype = 0, orig_qtype = 0, qclass = 0, name_len;
  u1 *po, *pi1;
  u2 *pi2;
  int got_soa = -1, got_ns = -1, too_big = 0;
  u1 *in_end = in_buf + packet_len;
  // Assume all queries on control sockets are ctl requests for now:
  int ctl_req = conn->ctl;
  int dnssec = 0;

  u2 flags = FB_QR | RC_NOERROR;
  int doopt = 0, didopt = 0;
  u1 optrcode = 0;
  u2 optflags = 0;
  int szlimit = (conn->sock->type[0] == 'T') ? MAX_TCP : (conn->sock->type[0] == 'U') ? MAX_UDP : MAX_MSG;
  u1 *out_end = out_buf + szlimit;
  memset (_out_buf, 0, 14);
  if (send_len)
    *send_len = 0;

  cinc_msg(packet_len);

  if (packet_len < 2)
  { // We haven't even got an ID
    logmsg(LOG_NOTICE, "(%s) invalid packet size: %i", constr(conn,0), packet_len);
    cinc_bad();
    return -1;
  }

  memset(_out_buf, 0, sizeof (_out_buf));
  po = out_buf + 12;
  pi2 = (u2 *)in_buf;
  id = ntohs (pi2[0]);

  if (conn->rcerr)
  { // We allready have an error code
    flags |= conn->rcerr;
    cinc_bad();
    goto empty_packet;
  }

  if (packet_len < 12) // Incomplete header
    packet_error(RC_FORMERR, RSP_NON, "(%s) invalid packet size: %i", constr(conn,0), packet_len);
  if (packet_len > MAX_MSG) // RFC-1035 says size limit of 512 - we make ours MAX_MSG for both UDP and TCP
    packet_error(RC_REFUSE, RSP_TCP, "(%s) packet too big: %i", constr(conn,0), packet_len);
  if (packet_len > MAX_BUF)
    packet_len = MAX_BUF;

  qflags = ntohs(pi2[1]);
  if ((qflags & FB_QR))	// response packet - ignoring
    packet_error(RC_FORMERR, RSP_NON, "(%s) response packet", constr(conn,0));

  // Copy flags for answer:
  flags |= qflags & (OPCODE | FB_RD); // Only opcode and recursion desired

  // get number of queries
  querycnt = ntohs(pi2[2]);
  if (querycnt < 1)
    packet_error(RC_FORMERR, RSP_TCP, "(%s) no queries", constr(conn,0));
#if !COUNTERS
  if (querycnt > 1)
    packet_error(RC_FORMERR, RSP_ALL, "(%s) %u queries", constr(conn,0), querycnt);
#endif

  // get the query string
  pi1 = in_buf + 12;
  memset(query, 0, sizeof (query));
  if (querycnt > 0)
  {
    if (get_name(in_buf, &pi1, in_buf + packet_len, query, sizeof (query), conn->ctl, 0) == -1)
      packet_error(RC_FORMERR, RSP_ALL, "(%s) bad query format", constr(conn,0));
    query_len = strlen ((char *) query);
    if (*query && query[query_len - 1] == '.')
      query[query_len - 1] = '\0';
  }

  // get the query type
  check_size(4, RSP_ALL, -1, 0);
  qtype = (qtype << 8) | *pi1++;
  qtype = (qtype << 8) | *pi1++;
  qclass = (qclass << 8) | *pi1++;
  qclass = (qclass << 8) | *pi1++;
  cinc_tym(qclass,qtype);

#if COUNTERS
  // Count types of additional queries (even though we won't answer them)
  for (j = 1; j < querycnt; j++)
  {
    u2 tqt = 0, tqc = 0;
    if (get_name(in_buf, &pi1, in_buf + packet_len, NULL, MAX_RECORD_LEN + 1, 0, 0) == -1)
      packet_error(RC_FORMERR, RSP_ALL, "(%s) bad label format", constr(conn,0));
    check_size(4, RSP_ALL, -1, j);
    tqt = (tqt << 8) | *pi1++;
    tqt = (tqt << 8) | *pi1++;
    tqc = (tqc << 8) | *pi1++;
    tqc = (tqc << 8) | *pi1++;
    cinc_tym(tqc,tqt);
  }
  if (querycnt > 1)
    packet_error(RC_FORMERR, RSP_ALL, "(%s) %u queries", constr(conn,0), querycnt);
#endif
  orig_qtype = qtype;

  u2 qmore[3];
  if ((qmore[EXTRAS] = ntohs(pi2[5])))
  {
    logmsg(LOG_DEBUG+2, "(%s) %u additional records", constr(conn,0), qmore[EXTRAS]);
#if !COUNTERS
    // Step through any additional queries unless we've allready done that:
    for (j = 1; j < querycnt; j++)
    {
      if (get_name(in_buf, &pi1, in_buf + packet_len, NULL, MAX_RECORD_LEN + 1, 0, 0) == -1)
        packet_error(RC_FORMERR, RSP_ALL, "(%s) bad label format", constr(conn,0));
      check_size(4, RSP_ALL, -1, j);
      pi1++; pi1++; // qtype
      pi1++; pi1++; // qclass
    }
#endif
    qmore[ANSWERS] = ntohs(pi2[3]);
    qmore[SERVERS] = ntohs(pi2[4]);
    u2 rdt = 0, rdu = 0, rdz = 0, rdl = 0;
    u1 rdf = 0, rdv = 0;
    for (j = 0; j < ANS_BLOCKS; j++)
      for (i = 0; i < qmore[j]; i++)
      {
        if ((k = get_name(in_buf, &pi1, in_buf + packet_len, NULL, MAX_RECORD_LEN + 1, 0, 0)) == -1)
          packet_error(RC_FORMERR, RSP_ALL, "(%s) bad label format", constr(conn,0));
        check_size(10, RSP_ALL, j, i);
        rdt = (rdt << 8) | *pi1++; // rrtype
        rdt = (rdt << 8) | *pi1++;
        rdu = (rdu << 8) | *pi1++; // rrclass / OPT UDP payload size
        rdu = (rdu << 8) | *pi1++;
        rdf = *pi1++; // OPT RCODE & flags
        rdv = *pi1++; // OPT Version
        rdz = (rdz << 8) | *pi1++; // OPT Z
        rdz = (rdz << 8) | *pi1++;
        rdl = (rdl << 8) | *pi1++; // rrlength
        rdl = (rdl << 8) | *pi1++;
        if ((j == EXTRAS) && (rdt == REQ_OPT))
        {
          cinc_opt(packet_len, rdu);
          doopt = 1;
          if (k != 0 )
            packet_error(RC_FORMERR, RSP_TCP, "(%s) bad OPT record", constr(conn,0));
          if ((rdz & OPTFB_BAD) != 0)
            packet_error(RC_NOTIMP, RSP_ALL, "(%s) Unsupported EDNS flag: %X", constr(conn,0), rdz);
          if (rdz & OPTFB_DO) {
            dnssec = dsec_on;
            cinc_sec();
          }
          logmsg(LOG_DEBUG+1, "(%s) got OPTv%u record; UDPz=%u, DO=%u, xRC=%u, xZ=%X, L=%u", constr(conn,0),
                 rdv, rdu, dnssec, rdf, rdz, rdl);
          if ((rdu > szlimit) && (szlimit < (name_len = ctl_req ? MAX_CTL : dnssec ? MAX_SEC : MAX_MSG)))
          {
            szlimit = (rdu > name_len) ? name_len : rdu;
            logmsg(LOG_DEBUG, "(%s) raised size limit: %i", constr(conn,0), szlimit);
          }
          else if ((conn->sock->type[0] == 'U') && (rdu < szlimit))
          {
            szlimit = rdu;
            logmsg(LOG_INFO, "(%s) lowered size limit: %i", constr(conn,0), szlimit);
          }
          out_end = out_buf + szlimit;
          break; // Only one OPT record
        }
        pi1 += rdl;
    }
  }

  // If there are more than one query, we only answer the first anyway
  nqueries = 1;

  // query section starts here ---->
  name_pointer = put_name(out_buf, &po, query, out_end);
  if (name_pointer < 0)
    packet_error(RC_SERVFAIL, RSP_ALL, "(%s) query too long", constr(conn,0));
  *po++ = qtype >> 8;
  *po++ = qtype & 0xFF;
  *po++ = qclass >> 8;
  *po++ = qclass & 0xFF;

  // Is this really a control request query?
  if (ctl_req)
  {
    if (sub_of_domain((char *)query, control_domain) < 0)
      ctl_req = 0;
    else
    {
      orig_qtype = 0;
      name_pointer = 0; // Answers need not contain full token domain
      logmsg(LOG_INFO, "(%s) Control request", constr(conn,0));
    }
  }

  if (qclass == CLASS_ANY)  // Any class?
    qclass = CLASS_INET;    // We only do INET
  else if (qclass != CLASS_INET) // not INET?
    packet_error(RC_NOTIMP, RSP_ALL, "(%s) unsupported query class: %u", constr(conn,0), qclass);
  // now check for valid flags. if not valid, jump over the part where we fill in the packet:
  if ((qflags & OPCODE) == OP_IQUERY) // inverse query
    packet_error(RC_NOTIMP, RSP_ALL, "(%s) unsupported inverse query", constr(conn,0)); // who supports this?
  if ((qflags & OPCODE) == OP_STATUS) // status request
    packet_error(RC_NOTIMP, RSP_ALL, "(%s) status request", constr(conn,0));
  if ((qflags & OPCODE) != OP_QUERY) // standard query?
    packet_error(RC_NOTIMP, RSP_ALL, "(%s) not standard query", constr(conn,0));

  // Setup query list
  lower_case_name(query);
  qsecs = add_qsec(NULL,NULL,ANS_BLOCKS,0,0,0,0,name_pointer,query);
  // Check for supported queries
  switch(qtype)
  {
    case REQ_ANY: // "any" request, send back what we have
      for (j = 0; j < SUPPORTED_RRS; j++)
      {
        if ((qtypes_supported[j] != REQ_RRSIG) && (!dnssec || (qtypes_supported[j] != REQ_NSEC)))
        { // except for NSEC and RRSIG if doing DNSSEC
          add_qsec(qsecs,NULL,ANSWERS,0,qtypes_supported[j],0,0,name_pointer,query);
          if (!dnssec)
            add_qsec(qsecs,NULL,ANSWERS,0,qtypes_supported[j],REQ_RRSIG,0,name_pointer,query);
        }
      }
      if (!conn->ctl)
      {
        add_qsec(qsecs,NULL,SERVERS,0,0,0,0,name_pointer,query);
        add_qsec(qsecs,NULL,EXTRAS,0,0,0,0,name_pointer,query);
      }
      break;
    case REQ_NS: // NS records makes us check NS before SOA in server section:
      add_qsec(qsecs,NULL,ANSWERS,0,qtype,0,0,name_pointer,query);
      if (!ctl_req)
      {
        add_qsec(qsecs,NULL,SERVERS,0,REQ_NS,0,0,name_pointer,query);
        add_qsec(qsecs,NULL,SERVERS,0,0,0,0,name_pointer,query);
        add_qsec(qsecs,NULL,EXTRAS,0,0,0,0,name_pointer,query);
      }
      break;
    case REQ_RRSIG: // Normally not queried for...
      for (j = 0; j < SUPPORTED_RRS; j++) // We need to check for RRSIGs for all supported types
        add_qsec(qsecs,NULL,ANSWERS,0,qtypes_supported[j],REQ_RRSIG,0,name_pointer,query);
      break;
    case REQ_OPT: // OPT pseudo records aren't normally queried for
      if (!conn->ctl) // opt records are not resource records:
        packet_error(0, RSP_ALL, "(%s) unsupported query type: %s", constr(conn,0), qtype_mname(qtype));
      // But on control sockets we send an empty response for them:
      break;
    default: // normal request:
      if (!(qtype_supported(qtype) || ((conn->ctl && (qtype_index(qtype) >= 0))))) // we don't serve this record type
        packet_error(0, RSP_ALL, "(%s) unsupported query type: %u %s", constr(conn,0), qtype, qtype_mname(qtype));
      add_qsec(qsecs,NULL,ANSWERS,0,qtype,0,0,name_pointer,query);
      if (!ctl_req)
      {
        add_qsec(qsecs,NULL,SERVERS,0,0,0,0,name_pointer,query);
        add_qsec(qsecs,NULL,EXTRAS,0,0,0,0,name_pointer,query);
      }
  }

  // answer section starts here ---->
  logmsg(LOG_INFO,"(%s) query: %s %s", constr(conn,0), qtype_mname(qtype), query);

  if (ctl_req || directory_check(REQ_SOA, query, DCL_DESCEND)) {
    logmsg(LOG_DEBUG+4,"(%s) authoritatively our: %s", constr(conn,0), query);
    flags |= FB_AA; // We're authoritative for this
  }
  if (dnssec)
  { // Set DNSSEC authetic data indicator?
    if (ctl_req)
    { // Answers to control requests are generated, so they're authentic
      flags |= FB_AD;
    }
    else if (ddns_domain_cmp((char *)query))
    { // Have we been assured that all dynamic data is authentic?
      if (dsec_ddns_auth)
        flags |= FB_AD;
    }
    else if (dsec_auth)
    { // We've been assured that all "static" data is authentic
      flags |= FB_AD;
    }
  }

  // Do we have *any* info on the domain?
  if (!ctl_req && !(conn->ctl && (qtype == REQ_PTR) && (*query == '*'))   && !directory_check(0, query, 0))
  {
    logmsg(LOG_DEBUG,"(%s) not found: %s", constr(conn,0), query);
    // No sense in trying to open files if the directory isn't there.
    for (qsec = qsecs; qsec; qsec = qsec->next)
      if (qsec->section == ANSWERS)
        qsec->type = 0;
    // if we know we are the authority, then we want to be explicit that the domain really does not exist:
    if (flags & FB_AA)
    {
      if (dnssec) // Add NSEC if DNSSEC
        add_qsec_NSEC(qsecs, query, FNS_WILD);
      logmsg(LOG_DEBUG+1,"(%s) authoritatively not found: %s", constr(conn,0), query);
      flags = (flags & ~RCODE) | RC_NXDOMAIN; // authoritatively no such domain
    }
  }

  for (j = 0, qsec = qsecs; qsec; j++, qsec = qsec->next)
  { // there are three answer blocks, but there may be multiple lookups per section
    // Prevent adding before this:
    qsec->prio = INT_MIN;
    if ((qsec->section >= ANS_BLOCKS) || !(qsec->section || qsec->type)) 
      continue; // Unused space in qsection list?

    // something replaced the query?
    if (qsec->nameptr)
      name_pointer = qsec->nameptr;
    else if (qsec->same && qsec->same->nameptr)
    {
      qsec->nameptr = qsec->same->nameptr;
      name_pointer = qsec->same->nameptr;
    }
    if (qsec->query)
    {
      strncpy((char *)query, (char *)qsec->query, sizeof(query));
      query[sizeof(query)-1] = 0;
    }
    // Unknown relation?
    if (qsec->reltype && !qsec->related)
    {
      if (!directory_check(qsec->type, query, DCL_MEMORY))
        continue;
      qsec->related = 1;
    }
    logmsg(LOG_DEBUG+1, "(%s) Section %d.%u.%d %u.%u", constr(conn,0), j, qsec->section+1, qsec->related, qsec->type, qsec->reltype);

    char **lookup_results = NULL;
    int added = 0;
    u2 *to_incr = NULL;
    char *ts1;
    int nosig = 0;
    if (qsec->related)
    { // Get record related to previous lookup:
      qtype = qsec->reltype ? qsec->reltype : qsec->type;
      lookup_results = related_lookup(qsec->type, qsec->reltype, query);
      if (!lookup_results)
        continue;
    }
    else if (ctl_req)
    { // Control channel request
      qtype = qsec->type;
      if (qsec->section != ANSWERS)
        continue;
      lookup_results = control_lookup(qtype);
      if (!lookup_results)
        continue;
    }
    else if (conn->ctl && (qsec->type == REQ_PTR) && (*query == '*'))
    { // Control socket request for host names
      qtype = qsec->type;
      lookup_results = names_lookup(qtype, query);
      if (!lookup_results)
        continue;
    }
    else if (qsec->type)
    { // Lookup the queried record:
      qtype = qsec->type;
      lookup_results = directory_lookup(qtype, query, DCL_MEMORY);
    }
    switch (qsec->section)
    { // Section based handling
      case ANSWERS:
        to_incr = &nanswers;
        if (!lookup_results && (orig_qtype != REQ_ANY))
        { // Check special replacements
          if (qsec->type == REQ_A || qsec->type == REQ_AAAA || qsec->type == REQ_A6) // Someone who asked for IPv6 might line IPv4 and vice versa:
            add_qsec_extra(qsecs, qsec->nameptr, qsec->query);
  	  if ((qsec->type != REQ_CNAME) && (qsec->type != REQ_NSEC))
  	  { // a missing a record may be a CNAME - try find it and then proceed as though we had a CNAME lookup:
            qtype = REQ_CNAME;
            lookup_results = directory_lookup(qtype, query, DCL_MEMORY);
  	  }
        }
        break;
      case SERVERS:
        to_incr = &nservers;
        if (!lookup_results && !qsec->type && got_soa)
        { // We haven't added an exact NS or SOA RR, so try to append here:
          i = directory_check(REQ_NS, query, DCL_DESCEND) ? found_cuts : -1;
          k = directory_check(REQ_SOA, query, DCL_DESCEND) ? found_cuts : -1;
          if ((k >= 0) && ((i < 0) || (i >= k)) && ((got_soa < 0) || (got_soa > k)) && ((got_ns < 0) || (got_ns > k)))
          { // We found an SOA RR that was a closer match than any SOA or NS RRs we've allready got:
            qtype = REQ_SOA;
            lookup_results = directory_lookup(qtype, query, DCL_DESCEND|DCL_MEMORY);
          }
          else if ((i >= 0) && ((got_soa < 0) || (got_soa > i)) && ((got_ns < 0) || (got_ns > i)))
          { // We found an NS RR that was a closer match than any SOA or NS RRs we've allready got:
            qtype = REQ_NS;
            lookup_results = directory_lookup(qtype, query, DCL_DESCEND|DCL_MEMORY);
          }
        }
        if (lookup_results && strncmp((char *)query, found_name, MAX_RECORD_LEN))
        {
          strncpy((char *)query, found_name, MAX_RECORD_LEN);
          name_pointer = -1;
        }
        if (dnssec && !lookup_results && (qtype == REQ_NSEC))
          logmsg(LOG_WARN,"No NSEC RR found for 's'", query);
        break;
      case EXTRAS:
        to_incr = &nextras;
        if (doopt && (put_OPT(out_buf, &po, optrcode, optflags, out_end) >= 0))
        { // Add EDNS OPT pseudo RR before other extras:
          doopt = 0;
          nextras++;
        }
        break;
    }

    // loop through each found record:
    for (i = 0; lookup_results && lookup_results[i] && i < 0x8000; i++)
    {
      u1 *saved_po = po;	        // push the offset in case we overrun or skip
      u1 *len_pointer = NULL;           // where length should be stored
      int skip_this = 0;                // flag to skip this record
      int auto_query = 0;               // flag if we want to lookup query again for extra section
      int auto_result = 0;              // flag if we want to lookup result for query section
      int only_one = 0;                 // flag if only one record is allowed in this answer
      int putres = 0;                   // name pointer to result, used for extras, negative means error
      int putres2 = 0;                  // name pointer to result, not used for extras, negative means error
      char *qres = lookup_results[i];   // pointer to result used for extras

      u4 ttl = TTL_POLICY((qsec->related && qsec->reltype) ? qsec->type : qtype, query);
      // pointer to original query string or CNAME result (2), or whatever the case may be:
      if (name_pointer < 1)
      { // Current query name not yet in packet
        name_pointer = put_name(out_buf, &po, query, out_end);
        if (name_pointer < 0)
          goto overflow_packet;
        if (name_pointer == 0)
        {
          qsec->nameptr = name_pointer;
          if (qsec->same)
            qsec->same->nameptr = name_pointer;
        }
      } else {
        *po++ = 0xc0 | ((name_pointer >> 8) & ~0xc0);
        *po++ = (name_pointer & 0xFF);
      }
      logmsg(LOG_DEBUG+1,"(%s) Result: %d.%u.%d.%d %s %s %s", constr(conn,0), j, qsec->section+1, qsec->related, i+1, query, qtype_mname(qtype), lookup_results[i]);
      *po++ = qtype >> 8;
      *po++ = qtype & 0xFF;
      *po++ = qclass >> 8;
      *po++ = qclass & 0xFF;
      *po++ = (ttl >> 24) & 0xFF;
      *po++ = (ttl >> 16) & 0xFF;
      *po++ = (ttl >> 8) & 0xFF;
      *po++ = (ttl >> 0) & 0xFF;
      len_pointer = po;	// store length of string here
      po += 2;

      switch(qtype)
      {
        case REQ_A:     
	  auto_query++;
          putres2 = put_A(out_buf, &po, lookup_results[i], out_end);
	  break;
        case REQ_AAAA:
	  auto_query++;
	  putres2 = put_AAAA(out_buf, &po, found_qtype, &qres, lookup_results[i], out_end);
	  break;
        case REQ_A6:
	  auto_query++;
	  putres = put_A6(out_buf, &po, found_qtype, &qres, lookup_results[i], out_end);
	  if (putres > 0)
	    auto_result++;
	  break;
        case REQ_SOA:
          putres = put_SOA(out_buf, &po, &qres, lookup_results[i], get_found_mtime(REQ_SOA, query), out_end);
	  if (putres > 0)
	  { 
	    if ((got_soa < 0) || (found_cuts < got_soa))
	      got_soa = found_cuts;
	    if (dnssec && ((qsec->section == ANSWERS) || (qsec->section == SERVERS)))
	    {
              add_qsec(qsecs, NULL, qsec->section, 0, REQ_DS, 0, 0, name_pointer, query);
              if (qsec->section == ANSWERS)
                add_qsec(qsecs, NULL, EXTRAS, 99, REQ_DNSKEY, 0, 0, name_pointer, query);
            }
            auto_result++;
          }
          only_one ++;	// force only one SOA record
	  break;
	case REQ_NS:    // Simple with extras
	  putres = put_name(out_buf, &po, (u1 *)lookup_results[i], out_end);
	  if (putres > 0)
	  {
	    if ((got_ns < 0) || (found_cuts < got_ns))
              got_ns = found_cuts;
	    if (dnssec && ((qsec->section == ANSWERS) || (qsec->section == SERVERS)))
            {
              add_qsec(qsecs, NULL, qsec->section, 0, REQ_DS, 0, 0, name_pointer, query);
              if (qsec->section == ANSWERS)
                add_qsec(qsecs, NULL, EXTRAS, 99, REQ_DNSKEY, 0, 0, name_pointer, query);
            }
            auto_result++;
          }
          if (!dnssec)
	    only_one ++;
	  break;
	case REQ_RRSIG:
          putres = put_RRSIG(out_buf, &po, &qres, qsec->type, lookup_results[i], out_end);
          if (dnssec && (putres > 0))
            add_qsec(qsecs, NULL, EXTRAS, 5, REQ_NS, 0, 0, putres, (u1 *)qres);
          auto_result++;
	  break;
        case REQ_MX:    
          putres = put_MX(out_buf, &po, lookup_results[i], i, out_end);
	  auto_result++;
	  break;
	case REQ_DNSKEY:
          putres2 = put_DNSKEY(out_buf, &po, lookup_results[i], out_end);
	  break;
        case REQ_SRV:   // for services, prefix with priority and weight (0) numbers
          putres = put_SRV(out_buf, &po, lookup_results[i], i, out_end);
	  auto_result++;
	  break;
        case REQ_SPF:  // SPF SHOULD be small, but we don't enforce that here
          putres2 = put_text(out_buf, &po, lookup_results[i], out_end);
	  break;
        case REQ_CNAME: // for cnames, we might need a special
          putres = put_name(out_buf, &po, (u1 *)lookup_results[i], out_end);
	  only_one ++;
	  if ((orig_qtype != REQ_CNAME) && (orig_qtype != REQ_ANY))
            add_qsec(qsecs,qsec,qsec->section,-10,qsec->type,0,0,putres,(u1 *)lookup_results[i]); // The original wasn't a CNAME so the client want's something else
	  else
	    auto_result++;
	  break;
        case REQ_TXT:  // for TXT store plain text in chunks of max 255 bytes
#if MACROS
          ts1 = expand_macro(lookup_results[i]);
          if (ts1)
            putres2 = put_text(out_buf, &po, ts1, out_end);
          else
#endif
            putres2 = put_text(out_buf, &po, lookup_results[i], out_end);
          break;
        case REQ_NSEC:
	  putres2 = put_NSEC(out_buf, &po, lookup_results[i], out_end);
	  only_one++;
	  break;
	case REQ_PTR:   // With extras
	  putres = put_name(out_buf, &po, (u1 *)lookup_results[i], out_end);
	  //only_one++;
	  auto_result++;
	  break;
	case REQ_DS:
          putres2 = put_DS(out_buf, &po, lookup_results[i], out_end);
	  break;
	case REQ_HINFO:
	  putres2 = put_HINFO(out_buf, &po, lookup_results[i], out_end);
	  break;
	default:        // don't store unknown types store string result
	  logmsg(LOG_DEBUG,"Unspecific answer qtype: %u %s",qtype,qtype_mname(qtype));
	  //putres = put_name(out_buf, &po, (u1 *)lookup_results[i], out_end);
	  skip_this++;
      }

      if ((putres < -2) || (putres2 < -2))
        skip_this++;
      else if ((putres < -1) || (putres2 < -1))
        format_error(qres);
      else if ((putres < 0) || (putres2 < 0) || (po > out_end))
      {	// limit to n bytes
       overflow_packet:
        skip_this++;
        too_big++;
        if ((qsec->section == ANSWERS) || ((qsec->section == SERVERS) && (qtype == REQ_RRSIG)))
        { // only answers and authority RRSIGs may set trunc flag
          flags |= FB_TC;
          logmsg(LOG_NOTICE,"Truncated: %d.%d %s %s", j, i, query, qtype_mname(qtype));
        }
        else
          logmsg(LOG_INFO,"Too big: %d.%d %s %s", j, i, query, qtype_mname(qtype));
        //po = po_within_n_bytes;
	//freen(lookup_results);
	//cinc_big();
	//goto full_packet;
      }

      putres2 = name_pointer;
      if (skip_this)
      { // don't add this record
       skip_record:
        skip_this ++;
        putres2 = 0;
        putres = 0;
      }

      // Do we want to add some extra stuff for this answer?
      if (!ctl_req && (qsec->section != EXTRAS) && (*query != '*'))
      {
        if (auto_query && ((putres2 > 0) || query[0]))  // For original query
          add_qsec_extra(qsecs, putres2, query);
	if (auto_result && ((putres > 0) || (qres && *qres))) // or result?
	  add_qsec_extra(qsecs, putres, (u1 *)qres);
      }

      if (skip_this || !len_pointer)
      { // don't add this record
        po = saved_po;
        continue;
      }

      if (only_one && lookup_results[i+1])
      { // force only one record
        nosig++; // Signature will be invalid if we remove RRs
        lookup_results[i+1] = NULL;
      }

      cinc_tyr(qtype);
      name_len = (po - len_pointer) - 2;
      *len_pointer++ = name_len >> 8;	// back insert the length
      *len_pointer++ = name_len & 0xFF;
      (*to_incr)++;
      added++;
    }
    if (dnssec && !ctl_req && ((qtype != REQ_RRSIG) || (orig_qtype == REQ_RRSIG)))
    { // If using DNSSEC, some RRs should be added
      if (lookup_results && added)
      {
        if (!nosig && (qtype != REQ_RRSIG)) // Add RR signatures:
          add_qsec(qsecs, qsec, qsec->section, -99, qtype, REQ_RRSIG, qsec->related+1, name_pointer, query);
        if ((found_name[0] == '*') && (qtype != REQ_NSEC) && (qsec->section == ANSWERS))
          add_qsec_NSEC(qsecs, query, 0); // And NSEC if exact match missing
      }
      else if ((qtype != REQ_NSEC) && (qsec->section == ANSWERS))
        add_qsec_NSEC(qsecs, query, FNS_WILD); // Add NSECs
    }
    freen(lookup_results);
  }
// full_packet:
  if (too_big) 
    cinc_big();
  while (qsecs) {
    qsec = qsecs;
    qsecs = qsecs->next;
    if (!qsec->same)
      freen(qsec->query);
    freen(qsec);
  }

 empty_packet:
  if (doopt && (put_OPT(out_buf, &po, optrcode, optflags, out_end) >= 0))
  { // Add EDNS OPT if not done
    doopt = 0;
    nextras++;
  }
  packet_len = (po - out_buf);
  po = out_buf;		// goto start of packet to fill in total counts
  *po++ = id >> 8;
  *po++ = id & 0xFF;
  *po++ = flags >> 8;
  *po++ = flags & 0xFF;
  *po++ = nqueries >> 8;
  *po++ = nqueries & 0xFF;
  *po++ = nanswers >> 8;
  *po++ = nanswers & 0xFF;
  *po++ = nservers >> 8;
  *po++ = nservers & 0xFF;
  *po++ = nextras >> 8;
  *po++ = nextras & 0xFF;

  // finally...
  cinc_res(packet_len, didopt);
  if (!rcode_ok(flags))
    cinc_err();
  else if (((flags & RCODE) == RC_NXDOMAIN) || !nanswers)
    cinc_non();
  logmsg(LOG_INFO,"(%s) answer: q=%u, a=%u, s=%u, e=%u, z=%u", constr(conn,0), nqueries, nanswers, nservers, nextras, packet_len);
  if (conn->sock->dgcn == SOCK_STREAM)
  { // TCP
    out_buf -= 2;
    out_buf[0] = (packet_len >> 8);
    out_buf[1] = (packet_len & 0xFF);
    packet_len += 2;
    if (send_buf && send_len)
    {
      k = sock_send(conn->sock, out_buf, packet_len, 0);
      if (k < packet_len)
      {
        if ((k == -1) && ((errno == EAGAIN) || (errno == EWOULDBLOCK)))
          k = 0;
        if (k >= 0)
        {
	  logmsg(LOG_DEBUG, "(%s) send %i/%u", constr(conn,0), k, packet_len);
          out_buf += k;
          packet_len -= k;
          *send_len = packet_len;
          *send_buf = realloc(*send_buf, *send_len);
          memcpy(*send_buf, out_buf, *send_len);
        }
      }
    }
    else
    {
      // this assumes the OS queue can hold our data without blocking. for 1282 bytes or less, we are pretty sure it can.
      k = sock_send(conn->sock, out_buf, packet_len, 1);
    }
  }
  else if (conn->sock->dgcn == SOCK_DGRAM)
  { // UDP
    k = sock_send_to(conn->sock, out_buf, packet_len);
  }
  else
    k = -1;
  return rcode_err(flags) ? -1 : k;
}

static void make_directories(uid_t usr, gid_t grp)
{
  logmsg(LOG_DEBUG+2, "data directory: %s", data_dir);
  if (!mkdir(data_dir, NEW_DIR_MODE))
    logmsg(LOG_NOTICE, "created data directory: %s", data_dir);
  char **hashes = hex_hashes(hash_type);
  if (hashes)
  {
    char dpath[264];
    strncpy(dpath, data_dir, 259);
    int dl = strlen(dpath);
    dpath[dl++] = '/';
    int dc = 0;
    for (int i = 0; hashes && hashes[i]; i++)
    {
      strncpy(dpath+dl, hashes[i], 4);
      if (!mkdir(dpath, NEW_DIR_MODE))
        dc++;
    }
    if (dc)
      logmsg(LOG_INFO, "created %i hash directories", dc);
    free(hashes);
  }
  make_directory("localhost", NULL);
  make_file("localhost", "A", "127.0.0.1", NULL);
  make_directory("localhost.localdomain", NULL);
  make_file("localhost.localdomain", "A", "127.0.0.1", NULL);
  make_directory("127.0.0.1", "localhost ipv4");
  make_file("127.0.0.1", "PTR", "localhost", "localhost ipv4 PTR");
  make_directory("0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.1", "localhost ipv6");
  make_file("0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.1","PTR", "localhost", "localhost ipv6 PTR");
}

void create_control_sockets()
{
  ip_socket *socks[8];
  int sockcnt = 8;
  int port = 0;
  if (sock_listen_random(socks, &sockcnt, &port))
    logerr("control socket not created");
  else
  {
    int ipv4udp = 0, ipv6udp = 0, ipv4tcp = 0, ipv6tcp = 0;
    for (int i = 0; i < sockcnt; i++)
    {
      struct interface *iface = malloc(sizeof (struct interface));
      iface->ctl = 1;
      iface->sock = socks[i];
      switch(socks[i]->ipv)
      {
        case AF_INET:
          switch(socks[i]->type[0])
          {
            case 'U':
              ipv4udp ++;
              break;
            case 'T':
              ipv4tcp ++;
              break;
          }
        case AF_INET6:
          switch(socks[i]->type[0])
          {
            case 'U':
              ipv6udp ++;
              break;
            case 'T':
              ipv6tcp ++;
              break;
          }
      }
      iface->next = interfaces.next;
      interfaces.next = iface;
    }
    if (control_domain)
      free(control_domain);
    control_domain = malloc(128);
    *control_domain = '.';
    random_name(control_domain+1, 128-10);
    strncpy(control_domain+(128-9), ".invalid", 9);
    logmsg(LOG_DEBUG, "control sockets created");
    char fn[1024];
    char buf[1024];
    if (snprintf(fn, sizeof(fn), "%s/%s", data_dir, ":control") < sizeof(fn))
    {
      snprintf(buf, sizeof(buf), "pid:%lld\n"
                                 "port:%d\n"
                                 "udpsz:%d\n"
                                 "tcpsz:%d\n"
                                 "ctlsz:%d\n"
                                 "domain:%s\n"
                                 "ipv4udp:%d\n"
                                 "ipv4tcp:%d\n"
                                 "ipv6udp:%d\n"
                                 "ipv6tcp:%d\n",
                                 (long long int)getpid(),
                                 port,MAX_MSG,MAX_TCP,MAX_CTL,control_domain,
                                 ipv4udp,ipv4tcp,ipv6udp,ipv6tcp);
      buf[sizeof(buf)-1] = 0;
      if (!write_file(fn, buf, S_IRUSR|S_IWUSR))
        logmsg(LOG_WARNING, "failed to create control file: %s", fn);
      else
      {
        chmod(fn, S_IRUSR|S_IWUSR);
        logmsg(LOG_DEBUG, "control file created: %s", fn);
      }
    }
  }
}

void create_sockets(char *ifaces)
{
  ip_socket *socks[8];
  int sockcnt;
  int totcnt = 0;
  char *service = "53";
  int i, idx = 0;
  char *ifstr;
  struct interface *iface;
  while ((ifstr = next_token(ifaces, &idx, '\t')))
  {
    switch(*ifstr)
    {
      case 'P':
        service = ifstr+1;
        break;
      case 'I':
        sockcnt = 8;
        if (!sock_listen(socks, &sockcnt, ifstr+1, service))
        {
          totcnt += sockcnt;
          for (i = 0; i < sockcnt; i++)
          {
            iface = malloc(sizeof (struct interface));
            iface->ctl = 0;
            iface->sock = socks[i];
            iface->next = interfaces.next;
            interfaces.next = iface;
          }
        }
        break;
    }
  }
  if (!totcnt)
    finish(1);
}

#define add_iface(w,s) do {             \
  r = ifaces ? strlen(ifaces) : 0;      \
  l = strlen(s);                        \
  ifaces = realloc(ifaces, r + l + 3);  \
  if (r) ifaces[r++] = '\t';            \
  ifaces[r++] = w;                      \
  strncpy(ifaces+r, s, l+1); } while (0)

int main(int argc, char **argv)
{
#if MAKEDBGD && !MAKEMEMW
  mtrace();
#endif
  program = DPROGRAM;
  versionstring = DVERSIONSTRING;
  usagestring = DUSAGESTRING;
  start_time = time(NULL);
#if COUNTERS
  memset(&counters, 0, sizeof(counters));
#endif
#if COUNTER_LOG
  int counter_log = COUNTER_LOG;
#endif
  int control_sockets = 0;
  srandom(time(NULL));

  int fork_twice = 0;
  char *pid_file = NULL;
  int maxconnections = MAX_CONNECTIONS;
  uid_t dusr = 0;
  gid_t dgrp = 0;
  if (!(getuid() && geteuid()))
  {
    dusr = get_user_id("nobody", dusr);
    dgrp = get_group_id("nogroup", get_group_id("nobody", dgrp));
  }
  char *ifaces = NULL;
  int *ifacec = 0;
  int r, l;

  read_config(INI_FILE);
  if (argc > 1)
  {
    int k;
    for (k = 1; k < argc; k++)
    {
      if (global_arg_option(argv[k]))
        continue;
      if ((k < argc - 1) && global_arg_setting(argv[k], argv[k+1]))
      {
        k++;
        continue;
      }

      if (!strncmp(argv[k], "-d", 3))
        fork_twice = 1;
      else if (!strncmp(argv[k], "-cs", 4))
        control_sockets = 1;
      else if (argv[k][0] == '-' && k > argc - 2)
      {
        fprintf(stderr, "%s\n", usagestring);
	finish(1);
      }
      else {
        string_chomp((u1 *)argv[k]);
        if (!strncmp(argv[k], "-u", 3))
          dusr = get_user_id(argv[++k], dusr);
        else if (!strncmp(argv[k], "-g", 3))
          dgrp = get_group_id(argv[++k], dgrp);
        else if (!strncmp(argv[k], "-ttl", 5))
          defttl_time = (u4)parse_duration(argv[++k]);
        else if (!strncmp(argv[k], "-attl", 6))
          srvttl_time = (u4)parse_duration(argv[++k]);
        else if (!strncmp(argv[k], "-ddttl", 6))
          dynttl_time = (u4)parse_duration(argv[++k]);
        else if (!strncmp(argv[k], "-p", 3))
        {
          k++;
          add_iface('P', argv[k]);
        }
        else if (!strncmp(argv[k], "-i", 3))
        {
          k++;
          add_iface('I', argv[k]);
          ifacec++;
        }
        else if (!strncmp(argv[k], "-pf", 4))
          pid_file = argv[++k];
        else
        {
          fprintf(stderr, "%s\n", usagestring);
          finish(1);
        }
      }
    }
  }

  logmsg(LOG_INFO,"%s starting", versionstring);
  make_directories(dusr, dgrp);

  u1 in_buf[MAX_BUF];
  if (srvttl_time < 1) srvttl_time = 60;
  if (defttl_time < 1) defttl_time = 30;
  if (dynttl_time < 1) dynttl_time = 10;

#if COUNTER_LOG
  int stosecs = (counter_log >= 600) ? 60 : (counter_log >= 60) ? 5 : 1;
#else
  int stosecs = 60;
#endif
  if (stosecs > MAX_CONN_TIME) stosecs = MAX_CONN_TIME;
  if (stosecs > MAX_IO_TIME)   stosecs = MAX_IO_TIME;
  if (stosecs > MAX_IDLE_TIME) stosecs = MAX_IDLE_TIME;

  struct sigaction sa;
  memset(&sa, 0, sizeof(sa));
  sa.sa_handler = interrupt;
  sigemptyset(&sa.sa_mask);
  sigaddset(&sa.sa_mask, SIGINT);
  sigaddset(&sa.sa_mask, SIGTERM);
  sigaddset(&sa.sa_mask, SIGPWR);
  sigaction(SIGINT, &sa, NULL);
  sigaction(SIGTERM, &sa, NULL);
  sigaction(SIGPWR, &sa, NULL);

  signal(SIGPIPE, SIG_IGN);
  signal(SIGHUP, SIG_IGN);
#ifdef SIGTSTP
  signal(SIGTSTP, SIG_IGN);
#endif
#ifdef SIGURG
  signal(SIGURG, SIG_IGN);
#endif

  if (!ifacec)
    add_iface('I', "");
  create_sockets(ifaces);
  free(ifaces);
  if (control_sockets)
    create_control_sockets();

  if (fork_twice)
  {
    if (fork() > 0)
      exit(0);
    if (fork() > 0)
      exit(0);
  }
  if (pid_file)
    write_pid(pid_file);

  if (dusr || dgrp)
  {
    if (geteuid() && !getuid())
      setuid(0);
    if (dgrp)
    {
      if (setgid(dgrp))
        logerr("set group id %d", dgrp);
      else
        logmsg(LOG_DEBUG,"changed group id: %d", dgrp);
    }
    if (dusr)
    {
      if (setuid(dusr))
        logerr("set user id %d", dusr);
      else
       logmsg(LOG_DEBUG,"changed user id: %d", dusr);
    }
  }

  logmsg(LOG_DEBUG,"ready");
  time_t current_time = time(NULL);
#if COUNTER_LOG
  time_t cntlogged_time = current_time;
#endif
  struct interface *iface;
  struct connection udpcon;
  memset(&udpcon, 0, sizeof(udpcon));
  while (1)
  {
    struct connection *c, *d;
    fd_set fdsetr;
    fd_set fdsetw;
    int f_max;
    struct timeval sto;

    // Set select data
    FD_ZERO(&fdsetr);
    FD_ZERO(&fdsetw);
    f_max = -1;
    // Add all interfaces
    for (iface = interfaces.next; iface;)
    { // Don't check TCP if we're busy
      if (sock_open(iface->sock) && ((iface->sock->dgcn != SOCK_STREAM) || (maxconnections <= 0) || (nconnections < maxconnections)))
      {
        FD_SET(iface->sock->fd, &fdsetr);
	f_max = max(f_max, iface->sock->fd);
      }
      iface = iface->next;
    }
    // Add all waiting connections
    for (c = &connections; c->next;)
    {
      d = c->next;
      if (abs(current_time - d->last_activity) > MAX_IDLE_TIME) // idle to long?
      {
        logmsg(LOG_DEBUG, "(%s) inactivity disconnect", constr(d,0));
        killtcp(c->next);
	continue;
      }
      f_max = max(f_max, d->sock->fd);
      switch(d->opmode)
      {
        case OPM_SEND: // Sending mode?
          if (abs(current_time - d->ioi_at) > MAX_IO_TIME) // Sending too long?
          {
	    logmsg(LOG_DEBUG, "(%s) slow output disconnect", constr(d,0));
            killtcp(c->next);
	    continue;
          }
	  FD_SET(d->sock->fd, &fdsetw);
        case OPM_RECV: // Receiving mode?
          if (abs(current_time - d->ioi_at) > MAX_IO_TIME) // Receiving too long?
          {
	    logmsg(LOG_DEBUG, "(%s) slow input disconnect", constr(d,0));
            killtcp(c->next);
	    continue;
          }
	  FD_SET(d->sock->fd, &fdsetr);
        default:  // Waiting mode
          if (abs(current_time - d->conn_at) > MAX_CONN_TIME) // Connected too long?
          {
	    logmsg(LOG_DEBUG, "(%s) long connection disconnect", constr(d,0));
            killtcp(c->next);
	    continue;
          }
	  FD_SET(d->sock->fd, &fdsetr);
      }
      c = c->next;
    }

    sto.tv_sec = stosecs;
    sto.tv_usec = 0;
    logmsg(LOG_DEBUG+5, "Wait (max %ds) for something...", stosecs);
    r = select(f_max + 1, &fdsetr, &fdsetw, NULL, &sto);
    if (r < 0)
    {
      logerr("select");
      break;
    }
    if (r == 0)
    {
      FD_ZERO(&fdsetr);
      FD_ZERO(&fdsetw);
    }

    current_time = time(NULL);

    // check all interfaces
    for (iface = interfaces.next; iface;)
    {
      if (sock_open(iface->sock) && FD_ISSET(iface->sock->fd, &fdsetr))
      {
        if (iface->sock->dgcn == SOCK_STREAM)
        { // Add a new TCP connection:
          ip_socket *sock = sock_accept(iface->sock);
          if (!sock)
      	    logerr("accept");
      	  else
          {
            c = (struct connection *)malloc(sizeof (struct connection));
      	    c->cid = ++idcounter;
      	    c->ctl = iface->ctl;
      	    c->sock = sock;
      	    c->ioi_at = c->conn_at = c->last_activity = current_time;
      	    c->opmode = OPM_WAIT;
      	    c->rcerr = RC_NOERROR;
      	    c->len = 0;
      	    c->alloced = 2;
      	    c->bufi = c->buf = malloc(c->alloced);
      	    c->next = connections.next;
      	    connections.next = c;
      	    nconnections++;
      	    cinc_tcp();
	    logmsg(LOG_INFO, "(%s) connect (%i)", constr(c,1), nconnections);
	  }
        }
        else if (iface->sock->dgcn == SOCK_DGRAM)
        { // Check for UDP packets
          if ((r = sock_receive_from(iface->sock, in_buf, sizeof(in_buf))) > 0)
	  {
	    udpcon.cid = ++idcounter;
	    udpcon.ctl = iface->ctl;
	    udpcon.sock = iface->sock;
	    udpcon.ioi_at = udpcon.conn_at = udpcon.last_activity = current_time;
      	    udpcon.rcerr = RC_NOERROR;
      	    cinc_udp();
      	    udpcon.sock->conn = 1;
	    process_packet(&udpcon, in_buf, r, NULL, NULL);
      	    udpcon.sock->conn = 0;
	  }
        }
      }  
      iface = iface->next;
    }

    // check all current TCP connections
    for (c = &connections; c->next;)
    {
      d = c->next;
      if (d->opmode == OPM_SEND)
      {
        if (FD_ISSET(d->sock->fd, &fdsetw))
        { // Try to send rest of data
          r = sock_send(d->sock, d->bufi, d->len, 0);
	  logmsg(LOG_DEBUG, "(%s) send %i/%u", constr(d,0), r, d->len);
          if ((r == -1) && (errno != EAGAIN) && (errno != EWOULDBLOCK))
          {
	    logmsg(LOG_DEBUG, "(%s) ouput lost connection", constr(d,0));
            killtcp(c->next);
            continue;
          }
          c->last_activity = current_time;
          if (r >= d->len)
          {
            if (d->rcerr < 0)
            {
              logmsg(LOG_DEBUG, "(%s) packet disconnect", constr(d,0));
              killtcp(c->next);
              continue;
            }
            d->opmode = OPM_WAIT;
            d->len = 0;
            d->alloced = 2;
            d->bufi = d->buf = realloc(d->buf, d->alloced);
          } else if (r > 0) {
            d->len -= r;
            d->bufi += r;
          }
        }
      }
      else
      if (FD_ISSET(d->sock->fd, &fdsetr))
      {	// got data?
        r = sock_receive(d->sock, d->buf + d->len, d->alloced - d->len, 0);
        if ((r == -1) && (errno == EAGAIN) && (errno == EWOULDBLOCK))
          r = 0;
        else if (r <= 0)
        {
          if (d->len == 0)
	    logmsg(LOG_DEBUG, "(%s) client disconnect", constr(d,0));
	  else
	    logmsg(LOG_DEBUG, "(%s) input lost connection", constr(d,0));
	  killtcp(c->next);
	  continue;
        }
        if (d->opmode != OPM_RECV)
        {
          d->rcerr = RC_NOERROR;
          d->ioi_at = current_time;
          d->opmode = OPM_RECV;
        }
	d->last_activity = current_time;
        if (r > 0)
        {
          d->bufi += r;
	  d->len += r;
	  if (d->len == 2) // first process the packet size bytes
	  {
	    u2 l;
	    l = (((u2)d->buf[0] << 8) | d->buf[1]);
	    if (l > MAX_MSG)
	    {
	      d->rcerr = RC_REFUSE;
	      logmsg(LOG_DEBUG, "(%s) size to high: %u", constr(d,0), l);
	      l = 12;
	      killtcp(c->next);
	      continue;
	    }
	    if (l == 0)
	    {
	      d->len = 0;
	      d->bufi = d->buf;
	      d->opmode = OPM_WAIT;
	    }
	    else
	    {
	      d->alloced = l + 2;	// add two bytes for the packet length
	      d->bufi = d->buf = realloc(d->buf, d->alloced);
	    }
	  }
	  else if (d->len == d->alloced)
	  {
            d->rcerr = process_packet(d, d->buf + 2, d->len - 2, &d->buf, &d->len);
            if (d->len)
            {
              d->ioi_at = current_time;
              d->opmode = OPM_SEND;
              d->alloced = d->len;
              d->bufi = d->buf;
            }
            else
            {
	      if (d->rcerr < 0)
	      {
	        logmsg(LOG_DEBUG, "(%s) packet disconnect", constr(d,0));
	        killtcp(c->next);
	        continue;
	      }
	      d->len = 0;
              d->opmode = OPM_WAIT;
	      d->alloced = 2;
	      d->bufi = d->buf = realloc(d->buf, d->alloced);
            }
	  }
        }
      }	 // ready for next packet
      c = c->next;
    }
#if COUNTER_LOG
    if (abs(current_time - cntlogged_time) >= counter_log)
    {
      log_counters();
      cntlogged_time = time(NULL);
    }
#endif
  }

  interrupt(0);
}
